#!/bin/bash
# v1.00: Initial version
# v1.05: Skip labels in export db when creating a spare vcartridge
# v1.06: Fix bug where label in exportdb can still be used

set -x

#some defaults
cachedir="/opt/quadstorvtl_impexp"
chunksize="10240m"
datefile=`date +%Y%m%d_%H%M%S`
createspare=1
compressdata=1
enckey="vtlkey.pem"

# Source rclone.conf to override defaults
if [ /quadstorvtl/etc/rclone.conf ]; then
	. /quadstorvtl/etc/rclone.conf
fi

#Validations

if [ "$cachedir" = "" ]; then
	echo "ERROR: cachedir not set"
	exit 22
fi

if [ "$1" = "%clear%" ]; then
	rm -rf $cachedir/*
	exit 0
fi

if [ "$chunksize" = "" ]; then
	echo "ERROR: chunksize not set"
	exit 22 
fi

#Read the arguments

vtl=$1
pool=$2
label=$3
fromimport=$4
postvaultlog="/quadstorvtl/tmp/postvault.log"

get_next_char() {
	asc=`printf "%d" "'$1"`
	if [ $asc -eq 57 ]; then
		nextchr="A"
		return
	elif [ $asc -ge 65 -a $asc -lt 90 ]; then
		let asc=asc+1
		nextchr=`printf "\x$(printf %x $asc)"`
		return
	elif [ $asc -ge 48 -a $asc -lt 57 ]; then
		let asc=asc+1
		nextchr=`printf "\x$(printf %x $asc)"`
		return
	elif [ $asc -ge 97 -a $asc -lt 122 ]; then
		let asc=asc+1
		nextchr=`printf "\x$(printf %x $asc)"`
		return
	else
		nextchr=""
	fi
}

gen_next_label() {
	vlabel=$1
	nextlabel=""
	suffix=""
	geo=""
	len=`expr length $vlabel`
	if [ $len -gt 2 ]; then
		geo=${vlabel:0:2}
	fi
	if [ $len -gt 6 ]; then
		suffix=${vlabel:6}
		vlabel=${vlabel:3:3}
		len=6
	fi

	tgt=""
	foundtgt=0
	for (( i=2; i>=0; i-- )); do
		chr=${vlabel:$i:1}
		get_next_char $chr
		if [ "$nextchr" != "" ]; then
			prefix=""
			if [ $i > 0 ]; then
				let idx=$i
				prefix=${vlabel:0:$idx}
			fi
			tgt="${prefix}${nextchr}${tgt}"
			foundtgt=1
			break;
		fi
		chr="0"
		tgt="${chr}${tgt}"
	done

	if [ $foundtgt -eq 1 ]; then
		nextlabel="${geo}V${tgt}${suffix}"
	fi
}

__create_spare_vcart() {
	vcartlabel=$1

	if [ "$createspare" = "0" ]; then
		return
	fi

	if [ -d /quadstorvtl/pgsqlsys ]; then
		vcartlist=`echo "select LABEL from vcartridge" | psql -t -p 9985 -U postgres qsdb`
		expvcartlist=`echo "select LABEL from expjobs" | psql -t -p 9985 -U postgres qsdb`
	else
		vcartlist=`echo "select LABEL from vcartridge" | /quadstorvtl/pgsql/bin/psql -t -U vtdbuser qsdb`
		expvcartlist=`echo "select LABEL from expjobs" | /quadstorvtl/pgsql/bin/psql -t -U vtdbuser qsdb`
	fi

	retries=0
	retriesmax=65536
	while [ $retries -lt $retriesmax ]; do
		gen_next_label $vcartlabel
		echo "For $vcartlabel got $nextlabel retries $retries" >> $postvaultlog
		if [ "$nextlabel" = "" ]; then
			break
		fi
		#Query DB to check if the next label is in the export db
		if [ -d /quadstorvtl/pgsqlsys ]; then
			sqlquery=`echo "select label from expjobs where UPPER(label)=UPPER('$nextlabel')" | psql -p 9985 -U postgres -t qsdb`
		else
			sqlquery=`echo "select label from expjobs where UPPER(label)=UPPER('$nextlabel')" | /quadstorvtl/pgsql/bin/psql -U vtdbuser -t qsdb`
		fi
		if [ "$sqlquery" != "" ]; then
			let retries=retries+1
			vcartlabel=$nextlabel
			continue
		fi

		isuniq=1
		for tmpcart in $vcartlist; do
			if [ "$tmpcart" = "$nextlabel" ]; then
				isuniq=0
				break
			fi
		done

		if [ $isuniq -eq 0 ]; then
			let retries=retries+1
			vcartlabel=$nextlabel
			continue
		fi

		for tmpcart in $expvcartlist; do
			if [ "$tmpcart" = "$nextlabel" ]; then
				isuniq=0
				break
			fi
		done

		if [ $isuniq -eq 1 ]; then
			break
		fi
		let retries=retries+1
		vcartlabel=$nextlabel
	done

	if [ "$nextlabel" = "" -o $retries -eq $retriesmax ]; then
		#Cannot find an unique label
		echo `date` "Cannot find an unique label for previous vcartridge ${label}" >> $postvaultlog
		return
	fi

	#Now we get the previous cartridge properties
	vcartrow=`/quadstorvtl/bin/vcconfig -l -v ${vtl} | sed -n '1!p' | grep " ${label} "`
	if [ "$vcartrow" = "" ]; then
		echo `date` "Cannot get properties for previous vcartridge ${label}" >> $postvaultlog
		return
	fi
	lto=`echo "$vcartrow" | awk '{print $5}'`

	if [ "$lto" != "LTO" ]; then
		echo `date` "Skipping non-LTO previous vcartridge ${label}" >> $postvaultlog
		return
	fi
	gen=`echo "$vcartrow" | awk '{print $6}'`
	vtype=0
	if [ $gen -eq 1 ]; then
		vtype=8
	elif [ $gen -eq 2 ]; then
		vtype=9
	elif [ $gen -eq 3 ]; then
		vtype=10
	elif [ $gen -eq 4 ]; then
		vtype=11
	elif [ $gen -eq 5 ]; then
		vtype=12
	elif [ $gen -eq 6 ]; then
		vtype=13
	elif [ $gen -eq 7 ]; then
		vtype=14
	fi

	if [ $vtype -eq 0 ]; then
		echo `date` "Cannot get type for previous vcartridge ${label}" >> $postvaultlog
		return
	fi

#	echo "next label $nextlabel for $label" >> $postvaultlog
	/quadstorvtl/bin/vcconfig -a -v $vtl -g $pool -p $nextlabel -t $vtype -c 1
	if [ $? -ne 0 ]; then
		echo `date` "Cannot exec /quadstorvtl/bin/vcconfig -a -v $vtl -g $pool -p $nextlabel -t $vtype -c 1" >> $postvaultlog
	else
		echo `date` "Created new vcartridge $nextlabel in VTL $vtl and pool $pool" >> $postvaultlog
		echo -n $nextlabel > /quadstorvtl/tmp/nextlabel
		logger `date` "Created new vcartridge $nextlabel in VTL $vtl and pool $pool"
	fi
}

create_spare_vcart() {
	echo `date` "Create spare cartridge for prev vcartridge $1" >> $postvaultlog
	logger `date` "Create spare cartridge for prev vcartridge $1"
	(
		flock -w 15 -e 200
		__create_spare_vcart $1
	) 200> /quadstorvtl/tmp/spare.lock
	echo `date` "Done create spare cartridge for prev vcartridge $1" >> $postvaultlog
	logger `date` "Done create spare cartridge for prev vcartridge $1"
}

#If the vcartridge was created from import and no changes to it, then delete it and exit
if [ "$fromimport" = "1" ]; then
	/quadstorvtl/bin/vcconfig -x -v $vtl -p $label
#	create_spare_vcart $label
	exit 0 
fi

keyarg=""
if [ "$enckey" != "" ]; then
	keyarg="--key=${enckey}"
fi

startlabel=$label
if [ -f /quadstorvtl/tmp/nextlabel ]; then
	startlabel=`cat /quadstorvtl/tmp/nextlabel`
fi
create_spare_vcart $startlabel
mkdir -p ${cachedir}/${label}
echo "`date` exec impexp for $label" >> $postvaultlog
logger "`date` exec impexp for $label"
echo "/quadstorvtl/bin/impexp -x -n -c $compressdata -e 1 -d ${cachedir}/${label} -f $label.img -p $label -v $vtl -s $chunksize -t $label-$datefile -k /quadstorvtl/scripts/cloudcopy.sh ${keyarg} -R /quadstorvtl/etc/rclone.conf" >> $postvaultlog
logger "/quadstorvtl/bin/impexp -x -n -c $compressdata -e 1 -d ${cachedir}/${label} -f $label.img -p $label -v $vtl -s $chunksize -t $label-$datefile -k /quadstorvtl/scripts/cloudcopy.sh ${keyarg} -R /quadstorvtl/etc/rclone.conf"
/quadstorvtl/bin/impexp -x -n -c $compressdata -e 1 -d ${cachedir}/${label} -f "$label.img" -p $label -v $vtl -s $chunksize -t "$label-$datefile" -k /quadstorvtl/scripts/cloudcopy.sh ${keyarg} -R /quadstorvtl/etc/rclone.conf >> $postvaultlog 2>&1
