#include "clstdefs.h"
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8)
#include <linux/blk-cgroup.h>
#endif
#include <commondefs.h>
#include <ioctldefs.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include "missingdefs.h"
#include "lz4.h"

static inline int
qs_deflate_block(uint8_t *in_buf, int uncomp_len, uint8_t *out_buf, int *comp_size, void *wrkmem, int algo)
{
	int retval;

	/* 4 bytes for comp size information */
	switch (algo) {
	case COMP_ALG_LZ4:
		retval = LZ4_compress_fast_extState(wrkmem, (char *)in_buf, (char *)(out_buf+sizeof(uint32_t)), uncomp_len, uncomp_len - 516, 1);
		break;
	default:
		retval = 0;
	}

	if (!retval)
		return -1;

	SET_COMP_SIZE(out_buf, retval, algo);
	*comp_size = (retval + sizeof(uint32_t));
	return 0;
}

static inline int
qs_inflate_block(uint8_t *in_buf, int comp_len, uint8_t *out_buf, int uncomp_len)
{
	int retval;
	uint32_t stored_comp_size, comp_size, algo;

	stored_comp_size = *((uint32_t *)(in_buf));
	comp_size = stored_comp_size & COMP_ALG_MASK;
	algo = stored_comp_size >> COMP_ALG_SHIFT;

	if (comp_size > comp_len)
		return -1;

	switch (algo) {
	case COMP_ALG_LZ4:
		retval = LZ4_decompress_safe((char *)(in_buf+sizeof(uint32_t)), (char *)out_buf, comp_size, uncomp_len);
		break;
	default:
		retval = 0;
	}

	if (retval != uncomp_len)
		return retval;
	return 0;
}
void __uma_zfree(uma_t *cachep, void *ptr);

static inline void
port_fill(uint64_t dest[], uint64_t src[])
{
	dest[0] = src[0];
	dest[1] = src[1];
}

static inline int
port_equal(uint64_t port1[], uint64_t port2[])
{
	return (port1[0] == port2[0] && port1[1] == port2[1]);
}

#define istate_mtx_free(x)	do {} while(0)
#define istate_wait_chan_free(x)	do {} while(0)
#define istate_uma_zfree		__uma_zfree
#define istate_mtx_alloc(xlck, nm)	spin_lock_init(&(xlck))
#define istate_mtx_lock(x)	spin_lock(&(x))
#define istate_mtx_unlock(x)	spin_unlock(&(x))
#define istate_chan_wakeup(chn) wake_up_all(&((chn).cv))	
#define istate_wait_chan_alloc(chn, nm) wait_chan_init(&(chn), nm)
#define tdevice_reservation_lock(tdk)		(mutex_lock(&(tdk)->reservation_lock))
#define tdevice_reservation_unlock(tdk)		(mutex_unlock(&(tdk)->reservation_lock))

#include "istate.h"

static sx_t cbs_lock;
static spinlock_t wrkmem_lock;
static uma_t *ctio_cache;
uma_t *istate_cache;
uma_t *node_sock_cache;
atomic_t itf_disabled;
static atomic_t qload_done;
struct interface_list cbs_list;
struct qs_interface_cbs *fc_icbs;
struct bdevice *bdevice;
static wait_chan_t bio_chan;
int msg_idx;
static spinlock_t bio_msg_lock;
wait_queue_head_t bio_msg_wait;
wait_queue_head_t comm_sock_wait;
static atomic_t core_inited;

struct wrkmem {
	uint8_t *wrkmem;
	void *pages;
	void *pglist;
	TAILQ_ENTRY(wrkmem) w_list;
};
TAILQ_HEAD(wrkmem_list, wrkmem);
struct wrkmem_list wrkmem_list = TAILQ_HEAD_INITIALIZER(wrkmem_list);
static atomic_t wrkmem_count;
uint8_t *msg_addr;

struct tdevice *device_list[MAX_DINFO_DEVICES];
static void tdevice_insert_ccb(struct qsio_hdr *ccb_h);

static void
tdevice_get(struct tdevice *tdevice)
{
	atomic_inc(&tdevice->refs);
}

static void
tdevice_put(struct tdevice *tdevice)
{
	struct initiator_state *istate;

	if (!atomic_dec_and_test(&tdevice->refs))
		return;
	while ((istate = SLIST_FIRST(&tdevice->istate_list)) != NULL) {
		SLIST_REMOVE_HEAD(&tdevice->istate_list, i_list);
		istate_put(istate);
	}
	kfree(tdevice);
}

pagestruct_t*
vm_pg_alloc(allocflags_t aflags)
{
	pagestruct_t *pp;
	int flags = GFP_NOIO | (aflags ? __GFP_ZERO : 0);

	pp = alloc_page(flags);
	return pp;
}

void
vm_pg_free(pagestruct_t *pp)
{
	__free_page(pp);
}

void*
vm_pg_address(pagestruct_t *pp)
{
	return (void *)(page_address(pp));
}

void
vm_pg_ref(pagestruct_t *pp)
{
	get_page(pp);
}

void
vm_pg_unref(pagestruct_t *pp)
{
	put_page(pp);
}

void
__qs_free(void *ptr, int flags)
{
	kfree(ptr);
}

void*
__qs_malloc(size_t size, int type, allocflags_t aflags)
{
	int flags = (aflags & Q_NOWAIT) ? GFP_ATOMIC : GFP_NOIO;
	int wait = (aflags & Q_WAITOK) ? 1 : 0;
	void *ret;

	while (!(ret = kmalloc(size, flags)) && wait)
		msleep(1);
	return ret;
}

void
__wait_complete_all(wait_compl_t *comp)
{
	unsigned long flags;

	mtx_lock_irqsave(&comp->mtx, flags);
	comp->done = 1;
	wake_up_all(&comp->cv);
	mtx_unlock_irqrestore(&comp->mtx, flags);
}

static void
cv_wait(wait_queue_head_t *cv, mtx_t *mtx, void *data, int intr)
{
	unsigned long *flags = data;
	DEFINE_WAIT(wait);

	add_wait_queue_exclusive(cv, &wait);
	if (!intr)
		set_current_state(TASK_UNINTERRUPTIBLE);
	else
		set_current_state(TASK_INTERRUPTIBLE);
	mtx_unlock_irqrestore(mtx, *flags);
	schedule();
	mtx_lock_irqsave(mtx, *flags);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(cv, &wait);
}

void
__wait_for_done(wait_compl_t *comp)
{
	unsigned long flags;

	mtx_lock_irqsave(&comp->mtx, flags);
	while (!comp->done) {
		cv_wait(&comp->cv, &comp->mtx, &flags, 0);
	}
	mtx_unlock_irqrestore(&comp->mtx, flags);

}

static long 
cv_timedwait(wait_queue_head_t *cv, mtx_t *mtx, void *data, int timo)
{
	unsigned long *flags = data;
	DEFINE_WAIT(wait);
	long ret;

	add_wait_queue_exclusive(cv, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
	mtx_unlock_irqrestore(mtx, *flags);
	ret = schedule_timeout(msecs_to_jiffies(timo));
	mtx_lock_irqsave(mtx, *flags);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(cv, &wait);
	return ret;
}

long 
__wait_for_done_timeout(wait_compl_t *comp, long timo)
{
	unsigned long flags;

	mtx_lock_irqsave(&comp->mtx, flags);
	while (!comp->done) {
		timo = cv_timedwait(&comp->cv, &comp->mtx, &flags, timo);
		if (!timo)
			break;
	}
	mtx_unlock_irqrestore(&comp->mtx, flags);
	return timo;
}

static void*
__uma_zcreate(const char *name, size_t size)
{
	void *cachep;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23))
	cachep = slab_cache_create(name, size, 0, 0, NULL, NULL);
#else
	cachep = slab_cache_create(name, size, 0, 0, NULL);
#endif
	return cachep;
}

static void
__uma_zdestroy(void *cachep)
{
	slab_cache_destroy(cachep);
}

void
__uma_zfree(uma_t *cachep, void *ptr)
{
	kmem_cache_free(cachep, ptr);
} 

void *
__uma_zalloc(uma_t *cachep, allocflags_t aflags)
{
	int flags = (aflags & Q_NOWAIT_INTR) ? GFP_ATOMIC : GFP_NOIO;
	int wait = (aflags & Q_WAITOK) ? 1 : 0;
	void *ret;

	while (!(ret = kmem_cache_zalloc(cachep, flags)) && wait)
		msleep(1);

	return ret;
}

void*
__qszalloc(size_t size, int type, allocflags_t aflags)
{
	int flags = (aflags & Q_NOWAIT_INTR) ? GFP_ATOMIC : GFP_NOIO;
	int wait = (aflags & Q_WAITOK) ? 1 : 0;
	void *ret;

	while (!(ret = kzalloc(size, flags)) && wait)
		msleep(1);
	return ret;
}

void
pgdata_free_page(struct pgdata *pgdata)
{
	if (pgdata->page) {
		vm_pg_free(pgdata->page);
		pgdata->page = NULL;
	}
}

int
pgdata_alloc_page(struct pgdata *pgdata, allocflags_t flags)
{
	if (pgdata->page)
		return 0;

	pgdata->page = vm_pg_alloc(flags);
	if (unlikely(!pgdata->page))
		return -1;
	return 0;
}

void
pgdata_copy_page_ref(struct pgdata *dest, pagestruct_t *page)
{
	memcpy(page_address(dest->page), vm_pg_address(page), LBA_SIZE);
}

void
pgdata_add_page_ref(struct pgdata *dest, pagestruct_t *page)
{
	vm_pg_ref(page);
	dest->page = page;
}

void
pgdata_copy_ref(struct pgdata *dest, struct pgdata *src)
{
	memcpy(page_address(dest->page), page_address(src->page), LBA_SIZE);
}

void
pgdata_add_ref(struct pgdata *dest, struct pgdata *src)
{
	pgdata_add_page_ref(dest, src->page);
	dest->pg_len = src->pg_len;
}

void
pgdata_move(struct pgdata *dest, struct pgdata *src)
{
	dest->page = src->page;
}

void
pglist_free(struct pgdata *pglist, int pglist_cnt)
{
	int i;

	for (i = 0; i < pglist_cnt; i++)
		pgdata_free_page(&pglist[i]);
	__qs_free(pglist, M_PGLIST);
}

void
pglist_free_nopage(struct pgdata *pglist, int pglist_cnt)
{
	__qs_free(pglist, M_PGLIST);
}

struct qsio_scsiio *
ctio_new(allocflags_t flags)
{
	return __uma_zalloc(ctio_cache, flags);
}

void
ctio_allocate_buffer(struct qsio_scsiio *ctio, int length, allocflags_t flags)
{
	int new_length;
	int pad_mask = (ctio->init_int != TARGET_INT_ISCSI) ? 0 : 0x03;

	if (pad_mask && (length & pad_mask)) {
		int pad = (pad_mask + 1) - (length & pad_mask);
		new_length = length + pad;
	} else
		new_length = length;
	ctio->data_ptr = __qs_malloc(new_length, M_CTIODATA, flags);
	ctio->dxfer_len = length;
}

void
tdevice_set_msg_pending(struct tdevice *tdevice)
{
	unsigned long flags;

	DEBUG_BUG_ON(atomic_read(&tdevice->pending_msg));
	spin_lock_irqsave(&tdevice->tdevice_lock, flags);
	atomic_set(&tdevice->pending_msg, 1);
	wake_up_interruptible(&tdevice->wait);
	spin_unlock_irqrestore(&tdevice->tdevice_lock, flags);
	wait_for_completion(&tdevice->ctio_completion);

}

static int
ctio_free_usr_pages_mmap(struct qsio_scsiio *ctio)
{
	struct tdevice *tdevice = ctio->ccb_h.tdevice;
	uint8_t *ptr = page_address(tdevice->ctio_page);
	struct ctio_free_req *req;
	struct core_node_msg *raw;

	raw = (struct core_node_msg *)(ptr);
	memset(raw, 0, sizeof(*req) + sizeof(*raw));
	raw->msg_cmd = NODE_MSG_CTIO_WRITE_CMD_FREE;
	raw->dxfer_len = sizeof(*req);

	req = (struct ctio_free_req *)(raw->data);
	memset(req, 0, sizeof(*req));
	req->lun_id = tdevice->lun_id;
	req->tx_id = ctio->tx_id;

	raw->msg_id = node_transaction_id();
	node_msg_compute_csum(raw);

	tdevice_set_msg_pending(tdevice);
	return 0;
}

static sx_t comm_socks_lock;
struct sys_sock_list comm_sock_list = TAILQ_HEAD_INITIALIZER(comm_sock_list);
struct sys_sock_list free_sock_list = TAILQ_HEAD_INITIALIZER(free_sock_list);

sock_t *
get_comm_sock(void)
{
	sock_t *sys_sock = NULL, *iter;
	DEFINE_WAIT(wait);
	int retries = 0;

again:
	sx_xlock(&comm_socks_lock);
	TAILQ_FOREACH(iter, &free_sock_list, f_entry) {
		if (sys_sock_is_connected(iter)) {
			TAILQ_REMOVE(&free_sock_list, iter, f_entry);
			sys_sock = iter;
			break;
		}
	}
	sx_xunlock(&comm_socks_lock);
	if (sys_sock)
		return sys_sock;
	add_wait_queue_exclusive(&comm_sock_wait, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
	schedule();
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&comm_sock_wait, &wait);
	if (retries++ < 30)
		goto again;
	DEBUG_WARN_NEW("Cannot get a free sock\n");
	return NULL;
}

void
put_comm_sock(sock_t *sys_sock)
{
	sx_xlock(&comm_socks_lock);
	TAILQ_INSERT_TAIL(&free_sock_list, sys_sock, f_entry);
	sx_xunlock(&comm_socks_lock);
	wake_up_interruptible(&comm_sock_wait);
}

static int
ctio_free_usr_pages(struct qsio_scsiio *ctio)
{
	struct tdevice *tdevice = ctio->ccb_h.tdevice;
	struct sys_sock *sock;
	struct ctio_free_req *req;
	struct core_node_msg *raw, raw_rsp;
	int retval;
	int error = -1;

	if (tdevice->ctio_page) {
		sx_xlock(&tdevice->ctio_lock);
		retval = ctio_free_usr_pages_mmap(ctio);
		sx_xunlock(&tdevice->ctio_lock);
		return retval;
	}

	sock = get_comm_sock();
	if (unlikely(!sock)) {
		debug_warn("cannot get a sock for cdb %x\n", ctio->cdb[0]);
		return -1;
	}

	raw = node_msg_alloc(sizeof(*req));
	memset(raw, 0, sizeof(*req) + sizeof(*raw));

	raw->msg_cmd = NODE_MSG_CTIO_WRITE_CMD_FREE;
	raw->dxfer_len = sizeof(*req);

	req = (struct ctio_free_req *)(raw->data);
	memset(req, 0, sizeof(*req));
	req->lun_id = tdevice->lun_id;
	req->tx_id = ctio->tx_id;

	retval = node_send_msg(sock, raw);
	if (unlikely(retval != 0))
		goto out;

	retval = node_sock_read(sock, &raw_rsp, sizeof(raw_rsp));
	if (unlikely(retval != 0))
		goto out;

	if (raw_rsp.msg_status != NODE_STATUS_OK)
		goto out;

	error = 0;
out:
	node_sock_finish(sock);
	node_msg_free(raw);
	return error;
}

void
ctio_free_data_nopage(struct qsio_scsiio *ctio)
{
	if (ctio->pglist_cnt)
		pglist_free_nopage((void *)ctio->data_ptr, ctio->pglist_cnt);
	else if (ctio->data_ptr)
		__qs_free(ctio->data_ptr, M_CTIODATA);
	ctio->data_ptr = NULL;
	ctio->dxfer_len = 0;
	ctio->pglist_cnt = 0;
}

void
ctio_free_data(struct qsio_scsiio *ctio)
{
	if (ctio->pglist_cnt && (ctio->ccb_h.flags & QSIO_USR_PAGES || (ctio->tx_id != 0))) {
		ctio_free_usr_pages(ctio);
		pglist_free_nopage((void *)ctio->data_ptr, ctio->pglist_cnt);
	}
	else if (ctio->pglist_cnt)
		pglist_free((void *)ctio->data_ptr, ctio->pglist_cnt);
	else
		__qs_free(ctio->data_ptr, M_CTIODATA);

	ctio->data_ptr = NULL;
	ctio->dxfer_len = 0;
	ctio->pglist_cnt = 0;
}

#define TDRIVE_STATS_ADD(tdrv,count,val)				\
do {									\
	if (tdrv)							\
		atomic64_add(val, (atomic64_t *)&tdrv->stats.count);		\
} while (0)

static void
copy_device_stats(struct vdeviceinfo *deviceinfo)
{
	struct tdevice *tdevice;
	uint32_t tid = (deviceinfo->lun_id - 1);

	if (tid >= MAX_DINFO_DEVICES)
		return;

	tdevice = device_list[tid];
	if (!tdevice)
		return;
	deviceinfo->stats.read_cmd_ticks = tdevice->stats.read_cmd_ticks;
	deviceinfo->stats.read_kernel_ticks = tdevice->stats.read_kernel_ticks;
	deviceinfo->stats.read_usr_ticks = tdevice->stats.read_usr_ticks;
	deviceinfo->stats.read_interface_ticks = tdevice->stats.read_interface_ticks;
	deviceinfo->stats.read_dxfer_ticks = tdevice->stats.read_dxfer_ticks;
	deviceinfo->stats.write_cmd_ticks = tdevice->stats.write_cmd_ticks;
	deviceinfo->stats.write_kernel_ticks = tdevice->stats.write_kernel_ticks;
	deviceinfo->stats.write_usr_ticks = tdevice->stats.write_usr_ticks;
	deviceinfo->stats.write_interface_ticks = tdevice->stats.write_interface_ticks;
	deviceinfo->stats.write_dxfer_ticks = tdevice->stats.write_dxfer_ticks;
	deviceinfo->stats.write_allocate_buffers_ticks = tdevice->stats.write_allocate_buffers_ticks;
}

static void
reset_device_stats(struct vdeviceinfo *deviceinfo)
{
	struct tdevice *tdevice;
	uint32_t tid = (deviceinfo->lun_id - 1);

	if (tid >= MAX_DINFO_DEVICES)
		return;

	tdevice = device_list[tid];
	if (!tdevice)
		return;
	memset(&tdevice->stats, 0, sizeof(tdevice->stats));
}

void
ctio_free_all(struct qsio_scsiio *ctio)
{
	struct tdevice *tdevice = ctio->ccb_h.tdevice;

	if (tdevice && (ctio->cdb[0] == READ_6 || ctio->cdb[0] == READ_16) && ctio->end_timestamp) {
		uint64_t kernel_ticks, usr_ticks, interface_ticks, cmd_ticks;

		kernel_ticks = ticks_delta(ctio->start_timestamp, ctio->queue_timestamp);
		kernel_ticks += ticks_delta(ctio->response_timestamp, ctio->end_timestamp);
		interface_ticks = ticks_delta(ctio->response_timestamp, ctio->end_timestamp);
		usr_ticks = ticks_delta(ctio->queue_timestamp, ctio->response_timestamp);
		cmd_ticks = ticks_delta(ctio->start_timestamp, ctio->end_timestamp);
		TDRIVE_STATS_ADD(tdevice, read_kernel_ticks, ticks_to_msecs(kernel_ticks));
		TDRIVE_STATS_ADD(tdevice, read_interface_ticks, ticks_to_msecs(interface_ticks));
		TDRIVE_STATS_ADD(tdevice, read_usr_ticks, ticks_to_msecs(usr_ticks));
		TDRIVE_STATS_ADD(tdevice, read_cmd_ticks, ticks_to_msecs(cmd_ticks));
		TDRIVE_STATS_ADD(tdevice, read_dxfer_ticks, ticks_to_msecs(ctio->dxfer_ticks));
	}
	else if (tdevice && (ctio->cdb[0] == WRITE_6 || ctio->cdb[0] == WRITE_16) && ctio->end_timestamp) {
		uint64_t kernel_ticks, usr_ticks, interface_ticks, cmd_ticks;

		kernel_ticks = ticks_delta(ctio->start_timestamp, ctio->queue_timestamp);
		kernel_ticks += ticks_delta(ctio->response_timestamp, ctio->end_timestamp);
		interface_ticks = ticks_delta(ctio->response_timestamp, ctio->end_timestamp);
		usr_ticks = ticks_delta(ctio->queue_timestamp, ctio->response_timestamp);
		cmd_ticks = ticks_delta(ctio->start_timestamp, ctio->end_timestamp);
		TDRIVE_STATS_ADD(tdevice, write_kernel_ticks, ticks_to_msecs(kernel_ticks));
		TDRIVE_STATS_ADD(tdevice, write_interface_ticks, ticks_to_msecs(interface_ticks));
		TDRIVE_STATS_ADD(tdevice, write_usr_ticks, ticks_to_msecs(usr_ticks));
		TDRIVE_STATS_ADD(tdevice, write_cmd_ticks, ticks_to_msecs(cmd_ticks));
		TDRIVE_STATS_ADD(tdevice, write_dxfer_ticks, ticks_to_msecs(ctio->dxfer_ticks));
		TDRIVE_STATS_ADD(tdevice, write_allocate_buffers_ticks, ticks_to_msecs(ctio->allocate_buffers_ticks));
	}

	ctio_free_data(ctio);
	ctio_free_sense(ctio);
	if (ctio->istate)
		istate_put(ctio->istate);
	__uma_zfree(ctio_cache, ctio);
}

static void
tdevice_ctio_reset(struct tdevice *tdevice)
{
	struct qsio_scsiio *ctio;

	spin_lock(&tdevice->ctio_alloc_lock);
	ctio = tdevice->ctio;
	tdevice->ctio = NULL;
	spin_unlock(&tdevice->ctio_alloc_lock);
	if (ctio)
		ctio_free_all(ctio);
}

static void
core_itf_proc_cmd(struct qsio_scsiio *ctio)
{
	struct tdevice *tdevice = ctio->ccb_h.tdevice;
	int retval;

	if (tdevice->ctio_page && (ctio->cdb[0] == WRITE_6 || ctio->cdb[0] == WRITE_16 || ctio->cdb[0] == READ_6 || ctio->cdb[0] == READ_16)) {
		sx_xlock(&tdevice->ctio_lock);
		retval = core_msg_send_ctio_mmap(ctio);
		sx_xunlock(&tdevice->ctio_lock);
	} else {
		if (ctio->cdb[0] == MODE_SELECT_6 || ctio->cdb[0] == MODE_SELECT_10)
			tdevice_ctio_reset(tdevice); /* only a mode select can change the write allocation characteristics */
		retval = core_msg_send_ctio(ctio);
	}
	if (retval != 0) {
		ctio_free_data(ctio);
		ctio_construct_sense(ctio, SSD_CURRENT_ERROR, SSD_KEY_HARDWARE_ERROR, 0, INTERNAL_TARGET_FAILURE_ASC, INTERNAL_TARGET_FAILURE_ASCQ);
	}
	ctio->response_timestamp = get_jiffies_64();
	device_send_ccb(ctio);
}

static int
pgdata_allocate_pgdata(struct qsio_scsiio *ctio, uint32_t block_size, uint32_t num_blocks, allocflags_t flags)
{
	struct pgdata *pglist;
	int pglist_cnt;

	pglist = pgdata_allocate(block_size, num_blocks, &pglist_cnt, Q_WAITOK, 0);
	if (unlikely(!pglist))
	{
		debug_warn("Allocation for pglist of num_blocks %u failed\n", num_blocks);
		return -1;
	}
	ctio->data_ptr = (void *)pglist;
	ctio->pglist_cnt = pglist_cnt;
	ctio->dxfer_len = block_size * num_blocks;
	return 0;
}

static int
device_allocate_buffers_mmap(struct qsio_scsiio *ctio, allocflags_t flags)
{
	struct tdevice *tdevice = ctio->ccb_h.tdevice;
	struct mmap_entry *ret = NULL;
	uint8_t *ptr = page_address(tdevice->ctio_page);
	struct ctio_allocate_req *req;
	struct ctio_allocate_rsp *rsp;
	struct core_node_msg *raw;
	struct pgdata_spec *spec;
	struct pgdata *pglist, *pgdata;
	uint8_t *spec_ptr = NULL;
	int spec_len, i;
	int error = -1;

	raw = (struct core_node_msg *)(ptr);
	memset(raw, 0, sizeof(*req) + sizeof(*raw));
	raw->msg_cmd = NODE_MSG_CTIO_WRITE_CMD_ALLOCATE;
	raw->dxfer_len = sizeof(*req);

	req = (struct ctio_allocate_req *)(raw->data);
	memcpy(req->cdb, ctio->cdb, 16);
	req->lun_id = tdevice->lun_id;

	raw->msg_id = node_transaction_id();
	node_msg_compute_csum(raw);

	tdevice_set_msg_pending(tdevice);

	if (raw->msg_status != NODE_STATUS_OK) {
		memset(raw, 0, sizeof(*raw));
		return -1;
	}

	rsp = (struct ctio_allocate_rsp *)(raw->data);

	pgdata_allocate_pgdata(ctio, rsp->block_size, rsp->num_blocks, Q_WAITOK);
	DEBUG_BUG_ON(!ctio->data_ptr);
	spec_len = (ctio->pglist_cnt * sizeof(*spec));

	if (raw->dxfer_len != (sizeof(*rsp) + spec_len)) {
		DEBUG_WARN_NEW("rsp dxfer len %d expected %d\n", raw->dxfer_len, (int)(sizeof(*rsp) + spec_len));
		goto out;
	}

	spec_ptr = (ptr + sizeof(*raw) + sizeof(*rsp));
	spec = (struct pgdata_spec *)spec_ptr;
	pglist = (struct pgdata *)ctio->data_ptr;
	ctio->ccb_h.flags |= QSIO_USR_PAGES;
	for (i = 0; i < ctio->pglist_cnt; i++, spec++) {
		pgdata = &pglist[i];
		pgdata->page = mmap_lookup_page((unsigned long)spec->addr, &ret);
		if (!pgdata->page) {
			DEBUG_WARN_NEW("Cannot lookup for addr %lu\n", spec->addr);
			goto out;
		}
		pgdata->pg_len = spec->pg_len;
		pgdata->pg_offset = spec->pg_offset;
	}
	ctio->ccb_h.flags |= QSIO_DATA_DIR_OUT;
	ctio->tx_id =  rsp->tx_id;
	error = 0;
out:
	if (error)
		ctio_free_data(ctio);
	memset(raw, 0, sizeof(*raw));
	return error;
}

static int
device_allocate_buffers(struct qsio_scsiio *ctio, allocflags_t flags)
{
	struct qsio_scsiio *saved_ctio;
	struct tdevice *tdevice = ctio->ccb_h.tdevice;
	struct mmap_entry *ret = NULL;
	sock_t *sock;
	struct ctio_allocate_req *req;
	struct ctio_allocate_rsp rsp;
	struct core_node_msg *raw, raw_rsp;
	struct pgdata_spec *spec;
	struct pgdata *pglist, *pgdata;
	uint8_t *spec_ptr = NULL;
	int retval, spec_len, i;
	int error = -1;

	spin_lock(&tdevice->ctio_alloc_lock);
	saved_ctio = tdevice->ctio;
	tdevice->ctio = NULL;
	spin_unlock(&tdevice->ctio_alloc_lock);

	if (saved_ctio) {
		if (memcmp(saved_ctio->cdb, ctio->cdb, 5) || ctio->cdb[0] != WRITE_6) {
			ctio_free_all(saved_ctio);
			goto alloc;
		}
		ctio->data_ptr = saved_ctio->data_ptr;
		ctio->dxfer_len = saved_ctio->dxfer_len; 
		ctio->pglist_cnt = saved_ctio->pglist_cnt;
		ctio->tx_id = saved_ctio->tx_id;
		ctio->ccb_h.flags |= QSIO_USR_PAGES;
		ctio->ccb_h.flags |= QSIO_DATA_DIR_OUT;
		__uma_zfree(ctio_cache, saved_ctio);
		return 0;
	}

alloc:
	if (tdevice->ctio_page) {
		sx_xlock(&tdevice->ctio_lock);
		retval = device_allocate_buffers_mmap(ctio, flags);
		sx_xunlock(&tdevice->ctio_lock);
		return retval;
	}

	sock = get_comm_sock();
	if (unlikely(!sock)) {
		debug_warn("cannot get a sock for cdb %x\n", ctio->cdb[0]);
		return -1;
	}

	raw = node_msg_alloc(sizeof(*req));
	memset(raw, 0, sizeof(*req) + sizeof(*raw));

	raw->msg_cmd = NODE_MSG_CTIO_WRITE_CMD_ALLOCATE;
	raw->dxfer_len = sizeof(*req);

	req = (struct ctio_allocate_req *)(raw->data);
	memcpy(req->cdb, ctio->cdb, 16);
	req->lun_id = tdevice->lun_id;

	retval = node_send_msg(sock, raw);
	if (unlikely(retval != 0))
		goto out;

	retval = node_sock_read(sock, &raw_rsp, sizeof(raw_rsp));
	if (unlikely(retval != 0))
		goto out;

	if (raw_rsp.msg_status != NODE_STATUS_OK)
		goto out;

	retval = node_sock_read(sock, &rsp, sizeof(rsp));
	if (unlikely(retval != 0))
		goto out;

	pgdata_allocate_pgdata(ctio, rsp.block_size, rsp.num_blocks, Q_WAITOK);
	DEBUG_BUG_ON(!ctio->data_ptr);
	spec_len = (ctio->pglist_cnt * sizeof(*spec));

	if (raw_rsp.dxfer_len != (sizeof(rsp) + spec_len)) {
		DEBUG_WARN_NEW("rsp dxfer len %d expected %d\n", raw_rsp.dxfer_len, (int)(sizeof(rsp) + spec_len));
		goto out;
	}

	spec_ptr = __qs_malloc(spec_len, M_MMAP, Q_WAITOK);

	retval = node_sock_read(sock, spec_ptr, spec_len);
	if (unlikely(retval != 0))
		goto out;

	spec = (struct pgdata_spec *)spec_ptr;
	pglist = (struct pgdata *)ctio->data_ptr;
	ctio->ccb_h.flags |= QSIO_USR_PAGES;
	for (i = 0; i < ctio->pglist_cnt; i++, spec++) {
		pgdata = &pglist[i];
		pgdata->page = mmap_lookup_page((unsigned long)spec->addr, &ret);
		if (!pgdata->page) {
			DEBUG_WARN_NEW("Cannot lookup for addr %lu\n", spec->addr);
			goto out;
		}
		pgdata->pg_len = spec->pg_len;
		pgdata->pg_offset = spec->pg_offset;
	}
	ctio->ccb_h.flags |= QSIO_DATA_DIR_OUT;
	ctio->tx_id =  rsp.tx_id;
	error = 0;
out:
	if (error)
		ctio_free_data(ctio);
	if (spec_ptr)
		free(spec_ptr, M_MMAP);
	node_sock_finish(sock);
	node_msg_free(raw);
	return error;
}

static int
ctio_write_length(struct qsio_scsiio *ctio, struct tdevice *tdevice, uint32_t *block_size, uint32_t *num_blocks, int *dxfer_len)
{
	sock_t *sock;
	struct ctio_write_len_req *req;
	struct core_node_msg *raw, raw_rsp;
	struct ctio_write_len_rsp rsp;
	int retval;
	int error = -1;

	sock = get_comm_sock();
	if (unlikely(!sock)) {
		debug_warn("cannot get a sock for cdb %x\n", ctio->cdb[0]);
		return -1;
	}

	raw = node_msg_alloc(sizeof(*req));
	memset(raw, 0, sizeof(*req) + sizeof(*raw));

	raw->msg_cmd = NODE_MSG_CTIO_WRITE_LENGTH;
	raw->dxfer_len = sizeof(*req);

	req = (struct ctio_write_len_req *)(raw->data);
	memset(req, 0, sizeof(*req));
	req->lun_id = tdevice->lun_id;
	memcpy(req->cdb, ctio->cdb, 16);

	retval = node_send_msg(sock, raw);
	if (unlikely(retval != 0))
		goto out;

	retval = node_sock_read(sock, &raw_rsp, sizeof(raw_rsp));
	if (unlikely(retval != 0))
		goto out;

	if (raw_rsp.msg_status != NODE_STATUS_OK)
		goto out;

	if (raw_rsp.dxfer_len != sizeof(rsp))
		goto out;

	retval = node_sock_read(sock, &rsp, sizeof(rsp));
	if (unlikely(retval != 0))
		goto out;

	*block_size = rsp.block_size;
	*num_blocks = rsp.num_blocks;
	*dxfer_len = rsp.dxfer_len;
	error = 0;
out:
	node_sock_finish(sock);
	node_msg_free(raw);
	return error;
}

static struct tdevice *
get_tdevice(uint32_t lun_id)
{
	if (lun_id >= MAX_DINFO_DEVICES)
		return NULL;
	return device_list[lun_id];
}

static inline void
device_set_vhba_id(struct tdevice *tdevice, int tid, int interface)
{
	if (interface == TARGET_INT_ISCSI)
		tdevice->iscsi_tid = tid;
	else if (interface == TARGET_INT_LOCAL)
		tdevice->vhba_id = tid;
}

static void
device_set_hpriv(struct tdevice *tdevice, void *hpriv)
{
	tdevice->hpriv = hpriv;
}

static uint32_t 
device_tid(struct tdevice *tdevice)
{
	return tdevice->lun_id;
}

static void
write_lun(struct tdevice *tdevice, uint8_t *ptr)
{
	__write_lun(tdevice->lun_id, ptr);
}

static int
device_queue_ctio(struct tdevice *tdevice, struct qsio_scsiio *ctio)
{
	ctio->queue_timestamp = get_jiffies_64();
	core_itf_proc_cmd(ctio);
	return 0;
}

static void
device_queue_ctio_list(struct ccb_list *ctio_list)
{
	struct qsio_hdr *ccb_h;
	struct qsio_scsiio *ctio;

	while ((ccb_h = STAILQ_FIRST(ctio_list)) != NULL) {
		STAILQ_REMOVE_HEAD(ctio_list, c_list);
		ctio = (struct qsio_scsiio *)ccb_h;
		if (ctio->init_int == TARGET_INT_FC)
			(*fc_icbs->ctio_exec)(ctio);
		else
			tdevice_insert_ccb(ccb_h);
	}
}

static struct pgdata *
pgdata_allocate_pglist(int nsegs, allocflags_t flags)
{
	struct pgdata *pglist;

	pglist = __qszalloc(sizeof(struct pgdata) * nsegs, M_PGLIST, flags);
	if (unlikely(!pglist)) {
		debug_warn("Allocation failure nsegs %d, flags %u\n", nsegs, flags);
		return NULL;
	}

	return pglist;
}

static int
pgdata_get_count(uint32_t block_size, uint32_t num_blocks)
{
	int pglist_cnt;

	if (!(block_size & LBA_MASK))
		pglist_cnt = ((block_size >> LBA_SHIFT)) * num_blocks;
	else
		pglist_cnt = ((block_size >> LBA_SHIFT) + 1) * num_blocks;

	return pglist_cnt;
}

struct pgdata *
pgdata_allocate(uint32_t block_size, uint32_t num_blocks, int *ret_pglist_cnt, allocflags_t flags, int alloc_page)
{
	struct pgdata *pglist, *pgtmp;
	int remaining;
	int pglist_cnt;
	int i, retval, idx;
	
	*ret_pglist_cnt = pglist_cnt = pgdata_get_count(block_size, num_blocks);
	debug_check(!pglist_cnt);

	pglist = pgdata_allocate_pglist(pglist_cnt, flags);
	if (unlikely(!pglist)) {
		debug_warn("allocating pglist for count %d failed\n", pglist_cnt);
		return NULL;
	}

	if (!alloc_page)
		return pglist;

	idx = 0;
	for (i = 0; i < num_blocks; i++) {
		for (remaining = block_size; remaining > 0; remaining -= LBA_SIZE) {
			pgtmp = &pglist[idx++];
			pgtmp->pg_len = (remaining > LBA_SIZE) ? LBA_SIZE : remaining;
			retval = pgdata_alloc_page(pgtmp, 0);
			if (unlikely(retval != 0)) {
				debug_warn("allocating for pgdata page failed\n");
				pglist_free(pglist, idx);
				return NULL;
			}
		}
	}
	return pglist;
}

static int
device_allocate_cmd_buffers(struct qsio_scsiio *ctio, allocflags_t flags)
{
	uint8_t *cdb = ctio->cdb;
	uint32_t parameter_list_length;

	switch (cdb[0])
	{
		case MODE_SELECT_6:
			parameter_list_length = cdb[4];
			break;
		case SEND_DIAGNOSTIC:
			parameter_list_length = be16toh(*(uint16_t *)(&cdb[3]));
			break;
		case MODE_SELECT_10:
		case PERSISTENT_RESERVE_OUT:
		case UNMAP:
			parameter_list_length = be16toh(*(uint16_t *)(&cdb[7]));
			break;
		case SPACE_16:
			parameter_list_length = be16toh(*(uint16_t *)(&cdb[12]));
			break;
		case WRITE_ATTRIBUTE:
			parameter_list_length = be32toh(*(uint32_t *)(&cdb[10]));
			if (parameter_list_length > 65536) /* Limit memory here */
				return -1;
			break;
		case SECURITY_PROTOCOL_OUT:
		case WRITE_DYNAMIC_RUNTIME_ATTRIBUTE:
			parameter_list_length = be32toh(*(uint32_t *)(&cdb[6]));
			if (parameter_list_length > 65536) /* Limit memory here */
				return -1;
			break;
		default:
			parameter_list_length = 0;
			debug_check(1);
	}

	debug_info("cdb %x parameter_list_length %d\n", cdb[0], parameter_list_length);
	ctio_allocate_buffer(ctio, parameter_list_length, flags);
	if (!ctio->data_ptr)
	{
		return -1;
	}
	ctio->ccb_h.flags |= QSIO_DATA_DIR_OUT;
	return 0;
}

static void
tdevice_insert_ccb(struct qsio_hdr *ccb_h)
{
	struct qsio_scsiio *ctio = (struct qsio_scsiio *)ccb_h;
	ctio->queue_timestamp = get_jiffies_64();
	core_itf_proc_cmd(ctio);
}

static int
ioctl_request(uint32_t cmd, void *send_buf, int send_len)
{
	sock_t *sock;
	struct ioctl_req *req;
	struct core_node_msg *raw, raw_rsp;
	int retval, error = -1;

	sock = get_comm_sock();
	if (unlikely(!sock)) {
		DEBUG_WARN_NEW("cannot get a sock for cmd %x\n", cmd);
		return -1;
	}

	raw = node_msg_alloc(sizeof(*req) + send_len);
	memset(raw, 0, sizeof(*raw));
	raw->msg_cmd = NODE_MSG_IOCTL;
	raw->dxfer_len = sizeof(*req) + send_len;

	req = (struct ioctl_req *)(raw->data);
	req->cmd = cmd;
	if (send_len)
		memcpy(req->data, send_buf, send_len);

	retval = node_send_msg(sock, raw);
	if (unlikely(retval != 0)) {
		DEBUG_WARN_NEW("node send msg failed for cmd %x retval %d\n", cmd, retval);
		goto out;
	}

	retval = node_sock_read(sock, &raw_rsp, sizeof(raw_rsp));
	if (unlikely(retval != 0)) {
		DEBUG_WARN_NEW("node sock read failed for cmd %x retval %d for sizeof raw_rsp\n", cmd, retval);
		goto out;
	}

	if (send_len) {
		if (raw_rsp.dxfer_len != send_len) {
			DEBUG_WARN_NEW("raw_rsp.dxfer_len %d send len %d\n", raw_rsp.dxfer_len, send_len);
			goto out;
		}
		retval = node_sock_read(sock, send_buf, send_len);
		if (unlikely(retval != 0)) {
			DEBUG_WARN_NEW("node sock read failed for cmd %x retval %d for send_len %d\n", cmd, retval, send_len);
			goto out;
		}
	}

	if (raw_rsp.msg_status != NODE_STATUS_OK) {
		switch (cmd) {
		case TLTARGIOCDDTABLELOADSTATUS:
			break;
		default:
			DEBUG_WARN_NEW("non zero msg status %d for cmd %u retval %d\n", raw_rsp.msg_status, _IOC_NR(cmd), retval);
			break;
		}
		goto out;
	}

	error = 0;
out:
	node_sock_finish(sock);
	node_msg_free(raw);
	return error;
}

static int
tdevice_request(struct tdevice *tdevice, uint64_t i_prt[], uint64_t t_prt[], uint8_t init_int, int msg_cmd, uint8_t *rsp, int rsp_len, uint32_t task_tag)
{
	sock_t *sock;
	struct initiator_req *req;
	struct core_node_msg *raw, raw_rsp;
	int retval, error = -1;

	sock = get_comm_sock();
	if (unlikely(!sock)) {
		debug_warn("cannot get a sock for cmd %x\n", msg_cmd);
		return -1;
	}

	raw = node_msg_alloc(sizeof(*req));
	memset(raw, 0, sizeof(*raw));
	raw->msg_cmd = msg_cmd;
	raw->dxfer_len = sizeof(*req);

	req = (struct initiator_req *)(raw->data);
	memset(req, 0, sizeof(*req));

	req->i_prt[0] = i_prt[0];
	req->i_prt[1] = i_prt[1];
	req->t_prt[0] = t_prt[0];
	req->t_prt[1] = t_prt[1];
	req->init_int = init_int;
	req->task_tag = task_tag;
	if (tdevice)
		req->lun_id = tdevice->lun_id;
	else
		req->lun_id = 0xFFFFFFFF;

	retval = node_send_msg(sock, raw);
	if (unlikely(retval != 0))
		goto out;

	retval = node_sock_read(sock, &raw_rsp, sizeof(raw_rsp));
	if (unlikely(retval != 0))
		goto out;

	if (raw_rsp.msg_status != NODE_STATUS_OK)
		goto out;

	if (rsp_len) {
		if (raw_rsp.dxfer_len != rsp_len)
			goto out;
		retval = node_sock_read(sock, rsp, rsp_len);
		if (unlikely(retval != 0))
			goto out;
	}

	error = 0;
out:
	node_sock_finish(sock);
	node_msg_free(raw);
	return error;
}

static int
tdevice_check_cmd(struct tdevice *tdevice, uint8_t op)
{
	uint64_t i_prt[2] = { 0, 0};

	return tdevice_request(tdevice, i_prt, i_prt, op, NODE_MSG_CHECK_CMD, NULL, 0, 0);
}

static int
device_istate_abort_task(struct tdevice *tdevice, uint64_t i_prt[], uint64_t t_prt[], int init_int, uint32_t task_tag)
{
	return tdevice_request(tdevice, i_prt, t_prt, init_int, NODE_MSG_ABORT_TASK, NULL, 0, task_tag);
}

static void
device_istate_abort_task_set(struct tdevice *tdevice, uint64_t i_prt[], uint64_t t_prt[], int init_int)
{
	tdevice_request(tdevice, i_prt, t_prt, init_int, NODE_MSG_ABORT_TASK_SET, NULL, 0, 0);
}

static int
tdevice_reset(struct tdevice *tdevice, uint64_t i_prt[], uint64_t t_prt[], uint8_t init_int)
{
	return tdevice_request(tdevice, i_prt, t_prt, init_int, NODE_MSG_DEVICE_RESET, NULL, 0, 0);
}

static void
device_free_initiator(uint64_t i_prt[], uint64_t t_prt[], int init_int, struct tdevice *tdevice)
{
	tdevice_request(tdevice, i_prt, t_prt, init_int, NODE_MSG_FREE_INITIATOR, NULL, 0, 0);
}

static void
device_new_initiator(uint64_t i_prt[], uint64_t t_prt[], int r_prt, int init_int, struct tdevice *tdevice, char *name)
{
	sock_t *sock;
	struct new_initiator_req *req;
	struct core_node_msg *raw, raw_rsp;
	int retval, error = -1;

	sock = get_comm_sock();
	if (unlikely(!sock)) {
		debug_warn("cannot get a sock\n");
		return;
	}

	raw = node_msg_alloc(sizeof(*req));
	memset(raw, 0, sizeof(*raw));
	raw->msg_cmd = NODE_MSG_NEW_INITIATOR;
	raw->dxfer_len = sizeof(*req);

	req = (struct new_initiator_req *)(raw->data);
	memset(req, 0, sizeof(*req));

	req->i_prt[0] = i_prt[0];
	req->i_prt[1] = i_prt[1];
	req->t_prt[0] = t_prt[0];
	req->t_prt[1] = t_prt[1];
	req->r_prt = r_prt;
	req->init_int = init_int;
	req->lun_id = tdevice->lun_id;
	if (name)
		strncpy(req->name, name, sizeof(req->name) -1);

	retval = node_send_msg(sock, raw);
	if (unlikely(retval != 0))
		goto out;

	retval = node_sock_read(sock, &raw_rsp, sizeof(raw_rsp));
	if (unlikely(retval != 0))
		goto out;

	if (raw_rsp.msg_status != NODE_STATUS_OK)
		goto out;

	error = 0;
out:
	node_sock_finish(sock);
	node_msg_free(raw);
}

int
fc_initiator_check(uint64_t wwpn[], uint64_t target_wwpn, void *device)
{
	return FC_RULE_ALLOW;
}

static int
__fc_initiator_check(uint64_t wwpn[], uint64_t target_wwpn, void *device)
{
	uint64_t t_prt[2];

	t_prt[0] = target_wwpn;
	t_prt[1] = 0;
	return tdevice_request(device, wwpn, t_prt, 0, NODE_MSG_FC_INITIATOR_CHECK, NULL, 0, 0);
}

static uint64_t
node_get_tprt(void)
{
        return 1;
}

int
vtdevice_register_interface(struct qs_interface_cbs *cbs)
{
	if (atomic_read(&itf_disabled))
		return -1;

	sx_xlock(&cbs_lock);
	switch (cbs->interface) {
	case TARGET_INT_LOCAL:
	case TARGET_INT_ISCSI:
	case TARGET_INT_FC:
		break;
	default:
		sx_xunlock(&cbs_lock);
		return -1;
	}

	if (!module_reference(THIS_MODULE)) {
		sx_xunlock(&cbs_lock);
		return -1;
	}

	cbs->ctio_new = ctio_new;
	cbs->ctio_allocate_buffer = ctio_allocate_buffer;
	cbs->ctio_free_data = ctio_free_data;
	cbs->ctio_free_all = ctio_free_all;
	cbs->ctio_write_length = ctio_write_length;
	cbs->get_device = get_tdevice;
	cbs->device_tid = device_tid;
	cbs->bus_from_lun = bus_from_lun;
	cbs->write_lun = write_lun;
	cbs->device_set_vhba_id = device_set_vhba_id;
	cbs->device_set_hpriv = device_set_hpriv;
	cbs->device_send_ccb = device_send_ccb;
	cbs->device_send_notify = device_send_notify;
	cbs->device_istate_abort_task = device_istate_abort_task;
	cbs->device_istate_abort_task_set = device_istate_abort_task_set;
	cbs->device_istate_queue_ctio = device_istate_queue_ctio;
	cbs->device_queue_ctio = device_queue_ctio;
	cbs->device_queue_ctio_list = device_queue_ctio_list;
	cbs->device_queue_ctio_direct = tdevice_insert_ccb;
	cbs->device_remove_ctio = device_remove_ctio;
	cbs->device_check_cmd = tdevice_check_cmd;
	cbs->device_allocate_buffers = device_allocate_buffers;
	cbs->device_allocate_cmd_buffers = device_allocate_cmd_buffers;
	cbs->device_target_reset = tdevice_reset;
	cbs->device_free_initiator = device_free_initiator;
	cbs->device_new_initiator = device_new_initiator;
	cbs->fc_initiator_check = __fc_initiator_check;
	cbs->get_tprt = node_get_tprt;
	cbs->qload_done = atomic_read(&qload_done);

	LIST_INSERT_HEAD(&cbs_list, cbs, i_list);
	if (cbs->interface == TARGET_INT_FC)
		fc_icbs = cbs;

	sx_xunlock(&cbs_lock);
	return 0;
}

static void
exit_caches(void)
{
	if (ctio_cache)
		__uma_zdestroy(ctio_cache);
	if (istate_cache)
		__uma_zdestroy(istate_cache);
	if (node_sock_cache)
		__uma_zdestroy(node_sock_cache);
}

static int
init_caches(void)
{
	ctio_cache = __uma_zcreate("vt_ctio_cache", sizeof(struct qsio_scsiio));
	if (!ctio_cache)
		return -1;

	istate_cache = __uma_zcreate("vt_istate_cache", sizeof(struct initiator_state));
	if (!istate_cache)
		return -1;

	node_sock_cache = __uma_zcreate("vt_node_sock_cache", sizeof(sock_t));
	if (!node_sock_cache)
		return -1;

	return 0;
}

void
vtdevice_unregister_interface(struct qs_interface_cbs *cbs)
{
	struct qs_interface_cbs *iter;
	int retval = -1;

	sx_xlock(&cbs_lock);
	LIST_FOREACH(iter, &cbs_list, i_list) {
		if (iter->interface != cbs->interface)
			continue;
		LIST_REMOVE(iter, i_list);
		if (iter->interface == TARGET_INT_FC)
			fc_icbs = NULL;
		retval = 0;
		break;
	}
	sx_xunlock(&cbs_lock);
	if (retval == 0)
		module_release(THIS_MODULE);
}

enum {
	DEVICE_ATTACHED,
};

static int
core_itf_device_new(struct ldev_info *ldev_info)
{
	struct qs_interface_cbs *cbs;
	struct tdevice *tdevice;
	uint32_t tid = (ldev_info->tid - 1);

	if (tid >= MAX_DINFO_DEVICES)
		return -1;

	if (device_list[tid])
		return -1;

	tdevice = kzalloc(sizeof(*tdevice), GFP_KERNEL);
	if (!tdevice)
		return -1;

	atomic_set(&tdevice->refs, 1);
	init_completion(&tdevice->ctio_completion);
	sx_init(&tdevice->ctio_lock, "ctio lock");
	spin_lock_init(&tdevice->tdevice_lock);
	spin_lock_init(&tdevice->ctio_alloc_lock);
	mutex_init(&tdevice->reservation_lock);
	SLIST_INIT(&tdevice->istate_list);
	init_waitqueue_head(&tdevice->wait);
	tdevice->type = TYPE_TDEVICE;
	tdevice->lun_id = ldev_info->tid;

	device_list[tid] = tdevice;
	sx_xlock(&cbs_lock);
	if (atomic_test_bit(DEVICE_ATTACHED, &tdevice->flags)) {
		sx_xunlock(&cbs_lock);
		ldev_info->iscsi_tid = tdevice->iscsi_tid;
		ldev_info->vhba_id = tdevice->vhba_id;
		ldev_info->lun_id = tdevice->lun_id;
		return 0;
	}

	LIST_FOREACH(cbs, &cbs_list, i_list) {
		(*cbs->new_device)(tdevice);
	}
	atomic_set_bit(DEVICE_ATTACHED, &tdevice->flags);
	sx_xunlock(&cbs_lock);
	ldev_info->iscsi_tid = tdevice->iscsi_tid;
	ldev_info->vhba_id = tdevice->vhba_id;
	ldev_info->lun_id = tdevice->lun_id;
	return 0;
}

static int
core_itf_device_delete(struct ldev_info *ldev_info)
{
	struct qs_interface_cbs *cbs;
	struct tdevice *tdevice;
	uint32_t tid = (ldev_info->tid - 1);

	if (tid >= MAX_DINFO_DEVICES)
		return -1;

	tdevice = device_list[tid];
	device_list[tid] = NULL;
	if (!tdevice) {
		DEBUG_INFO_NEW("Cannot find device at %d\n", tid);
		return -1;
	}

	sx_xlock(&cbs_lock);
	LIST_FOREACH(cbs, &cbs_list, i_list) {
		if (cbs->interface == TARGET_INT_ISCSI)
			(*cbs->remove_device)(tdevice, tdevice->iscsi_tid, tdevice->hpriv);
		else
			(*cbs->remove_device)(tdevice, tdevice->vhba_id, tdevice->hpriv);
	}
	sx_xunlock(&cbs_lock);
	wake_up_interruptible(&tdevice->wait);

	tdevice_put(tdevice);
	return 0;
}

static void
core_itf_detach_interfaces(void)
{
	struct qs_interface_cbs *iter;
	struct qs_interface_cbs *fc;

again:
	fc = NULL;
	sx_xlock(&cbs_lock);
	LIST_FOREACH(iter, &cbs_list, i_list) {
		if (!iter->detach_interface)
			continue;
		fc = iter;
		break;
	}
	sx_xunlock(&cbs_lock);
	if (!fc)
		return;

	(*fc->detach_interface)();
	goto again;
}

static int
core_itf_device_disable(struct ldev_info *ldev_info)
{
	struct qs_interface_cbs *cbs;
	struct tdevice *tdevice;
	uint32_t tid = (ldev_info->tid - 1);

	if (tid >= MAX_DINFO_DEVICES)
		return -1;

	tdevice = device_list[tid];
	if (!tdevice)
		return -1;

	tdevice_ctio_reset(tdevice);
	atomic_set(&tdevice->disabled, 1);
	wake_up_interruptible(&tdevice->wait);

	sx_xlock(&cbs_lock);
	LIST_FOREACH(cbs, &cbs_list, i_list) {
		if (!cbs->disable_device)
			continue;
		if (cbs->interface == TARGET_INT_ISCSI)
			(*cbs->disable_device)(tdevice, tdevice->iscsi_tid, tdevice->hpriv);
		else
			(*cbs->disable_device)(tdevice, tdevice->vhba_id, tdevice->hpriv);
	}
	sx_xunlock(&cbs_lock);
	return 0;
}

static int
core_itf_device_update(struct ldev_info *ldev_info)
{
	struct qs_interface_cbs *cbs;
	struct tdevice *tdevice;
	uint32_t tid = (ldev_info->tid - 1);

	if (tid >= MAX_DINFO_DEVICES)
		return -1;

	tdevice = device_list[tid];
	if (!tdevice)
		return -1;

	sx_xlock(&cbs_lock);
	LIST_FOREACH(cbs, &cbs_list, i_list) {
		if (!cbs->update_device)
			continue;
		if (cbs->interface == TARGET_INT_ISCSI)
			(*cbs->update_device)(tdevice, tdevice->iscsi_tid, tdevice->hpriv);
		else
			(*cbs->update_device)(tdevice, tdevice->vhba_id, tdevice->hpriv);
	}
	sx_xunlock(&cbs_lock);

	return 0;
}

#define MMAP_TABLE_ENTRIES		(1024 * 1024)

struct mmap_table {
	struct mmap_entry_list entry_list;
	mtx_t table_lock;
};

struct mmap_table mmap_table[MMAP_TABLE_ENTRIES];

static void
mmap_table_exit(void)
{
	struct mmap_entry *entry;
	struct mmap_table *table;
	int i;

	for (i = 0; i < MMAP_TABLE_ENTRIES; i++) {
		table = &mmap_table[i];
		while ((entry = TAILQ_FIRST(&mmap_table->entry_list)) != NULL) {
			TAILQ_REMOVE(&mmap_table->entry_list, entry, e_list);
			free(entry, M_MMAP);
		}
	}
}

static void
mmap_table_init(void)
{
	struct mmap_table *table;
	int i;

	for (i = 0; i < MMAP_TABLE_ENTRIES; i++) {
		table = &mmap_table[i];
		spin_lock_init(&table->table_lock);
		TAILQ_INIT(&table->entry_list);
	}
}

static void
mmap_entry_add(unsigned long start_addr, unsigned long end_addr, pagestruct_t *page)
{
	struct mmap_entry *entry;
	struct mmap_table *table;
	unsigned long idx;

	entry = __qszalloc(sizeof(*entry), M_MMAP, Q_WAITOK);
	entry->start_addr = start_addr;
	entry->end_addr = end_addr;
	entry->page = page;

	idx = (start_addr >> MAP_ALLOCATION_SHIFT);
	idx = idx % (MMAP_TABLE_ENTRIES - 1);
	table = &mmap_table[idx];

	mtx_lock(&table->table_lock);
	TAILQ_INSERT_TAIL(&table->entry_list, entry, e_list);
	mtx_unlock(&table->table_lock);
//	DEBUG_INFO_NEW("start addr %lu page %p shift %lu idx %lu\n", start_addr, page, (start_addr >> MAP_ALLOCATION_SHIFT), idx);
}

static void
mmap_entry_remove(unsigned long start_addr)
{
	struct mmap_entry *entry;
	struct mmap_table *table;
	unsigned long idx;

	idx = (start_addr >> MAP_ALLOCATION_SHIFT);
	idx = idx % (MMAP_TABLE_ENTRIES - 1);
	table = &mmap_table[idx];

	mtx_lock(&table->table_lock);
	TAILQ_FOREACH(entry, &table->entry_list, e_list) {
		if (entry->start_addr != start_addr)
			continue;
		TAILQ_REMOVE(&table->entry_list, entry, e_list);
	//	DEBUG_INFO_NEW("start addr %lu shift %lu idx %lu\n", start_addr, (start_addr >> MAP_ALLOCATION_SHIFT), idx);
		free(entry, M_MMAP);
		break;
	}
	mtx_unlock(&table->table_lock);
}

pagestruct_t *
mmap_lookup_page(unsigned long addr, struct mmap_entry **ret)
{
	struct mmap_entry *entry = *ret;
	struct mmap_table *table;
	pagestruct_t *base = NULL;
	unsigned long offset = 0;
	unsigned long idx, start_idx;
	int done_neg = 0, done_pos = 0;

	if (entry && addr >= entry->start_addr && addr < entry->end_addr) {
		base = entry->page;
		offset = (addr - entry->start_addr);
		goto out;
	}

	idx = (addr >> MAP_ALLOCATION_SHIFT);
	idx = idx % (MMAP_TABLE_ENTRIES - 1);
	start_idx = idx;

retry:
	table = &mmap_table[idx];
	mtx_lock(&table->table_lock);
	TAILQ_FOREACH(entry, &table->entry_list, e_list) {
		if (addr >= entry->start_addr && addr < entry->end_addr) {
			base = entry->page;
			offset = (addr - entry->start_addr);
			*ret = entry;
			break;
		}
	}
	mtx_unlock(&table->table_lock);

	if (!base) {
		if (done_neg && done_pos) {
			return NULL;
		}
		if (!done_neg) {
			done_neg = 1;
			idx--;
			if (idx == 0)
				idx = (MMAP_TABLE_ENTRIES - 1);
		}
		else {
			done_pos = 1;
			idx = start_idx + 1;
			if (idx == (MMAP_TABLE_ENTRIES - 1))
				idx = 0;
		}
		goto retry;
	}
out:
	if (!offset)
		return base;
	addr = (unsigned long)page_address(base) + offset;
	return virt_to_page(addr);
}

struct mmap_priv {
	struct page *page;
	unsigned long start_addr;
	unsigned long size;
};

static void
mmap_vma_close(struct vm_area_struct *vma)
{
	struct mmap_priv *priv = (struct mmap_priv *)vma->vm_private_data;
	unsigned long i, count, offset = 0, addr;
	pagestruct_t *page;

	page = priv->page;
	count = (priv->size >> PAGE_SHIFT);
	for (i = 0; i < count; i++) {
		addr = (unsigned long)page_address(page) + offset;
		__free_page(virt_to_page(addr));
		offset += PAGE_SIZE;
	}
	mmap_entry_remove(priv->start_addr);
	kfree(priv);
	vma->vm_private_data = NULL;
}

static struct vm_operations_struct mmap_remap_vm_ops = {
	.close = mmap_vma_close,
};

static int
coreitf_mmap(struct file *filp, struct vm_area_struct *vma)
{
	unsigned long size = vma->vm_end - vma->vm_start;
	unsigned long order;
	struct page *page;
	struct mmap_priv *priv;
	int retval;

	priv = kmalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		return -ENOMEM;
	}

	order = get_order(size);
	page = alloc_pages(GFP_KERNEL, order);
	if (!page) {
		kfree(priv);
		DEBUG_WARN_NEW("Could not alloc page for size %lu order %lu\n", size, (unsigned long)order);
		return -ENOMEM;
	}

	retval = remap_pfn_range(vma, vma->vm_start, page_to_pfn(page), size, vma->vm_page_prot);
	if (retval < 0) {
		__free_pages(page, order);
		kfree(priv);
		DEBUG_WARN_NEW("Could not remap page for size %lu\n", size);
		return retval;
	}
	if (size != PAGE_SIZE)
		split_page(page, order);
	priv->page = page;
	priv->size = size;
	priv->start_addr = vma->vm_start;
//	DEBUG_INFO_NEW("vma start %lu %p after map %p pgaddr %lu %lu\n", (unsigned long)vma->vm_start, virt_to_page(vma->vm_start), page, (unsigned long)page_address(virt_to_page(vma->vm_start)), (unsigned long)page_address(page));
	vma->vm_ops = &mmap_remap_vm_ops;
	vma->vm_private_data = priv;
	mmap_entry_add(vma->vm_start, vma->vm_end, page);
	return 0;
}

#define TL_MAX_KFDS	4096
static iodev_t *bdevs[TL_MAX_KFDS];
static struct scsi_device *tdevs[TL_MAX_KFDS];
static sx_t bdevs_lock;
static sx_t tdevs_lock;
static sx_t socks_lock;
#define TL_MAX_SOCKS	65536
static sock_t *socks[TL_MAX_SOCKS];

extern int validate_block_device(const char *devpath,  iodev_t *b_dev);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
static void
core_itf_iodev_close(iodev_t *b_dev)
{
	if (b_dev)
		fput(b_dev);
}
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(6,5,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR == 4) || (defined(CONFIG_SUSE_KERNEL) && LINUX_VERSION_CODE >= KERNEL_VERSION(6, 4, 0))
static void
core_itf_iodev_close(iodev_t *b_dev)
{
	if (b_dev)
		blkdev_put(b_dev, THIS_MODULE);
}
#else
static void
core_itf_iodev_close(iodev_t *b_dev)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29))
	int flags = FMODE_READ | FMODE_WRITE;
#endif 

	if (!b_dev)
		return;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0))
	blkdev_put(b_dev, flags);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29))
	close_bdev_exclusive(b_dev, flags);
#else
	close_bdev_excl(b_dev);
#endif  
}
#endif

#define MAX_SEND_THREADS	512

static void
core_send_exit(void)
{
	sock_t *sys_sock;

	while ((sys_sock = TAILQ_FIRST(&comm_sock_list)) != NULL) {
		TAILQ_REMOVE(&comm_sock_list, sys_sock, l_entry);
		sys_sock_close(sys_sock, 0);
		sys_sock_free(sys_sock);
	}
}

static int
core_send_init(void)
{
	int i, retval;
	sock_t *sys_sock;

	for (i = 0; i < MAX_SEND_THREADS; i++) {
		sys_sock = __sys_sock_create(NULL, AF_LOCAL);
		if (!sys_sock)
			return -ENOMEM;

		retval = sys_sock_connect_local(sys_sock, CDAEMON_RECV_PATH);
		if (unlikely(retval != 0)) {
			debug_warn("sock connect failed for %d retval %d\n", i, retval);
			sys_sock_free(sys_sock);
			return retval;
		}
		sx_xlock(&comm_socks_lock);
		TAILQ_INSERT_TAIL(&comm_sock_list, sys_sock, l_entry);
		TAILQ_INSERT_TAIL(&free_sock_list, sys_sock, f_entry);
		sx_xunlock(&comm_socks_lock);

		if ((i % 8) == 0)
			pause("psg", 50);
	}
	return 0;
}

static sock_t *
get_sys_sock(int sockfd)
{
	sock_t *sys_sock;

	if (sockfd >= TL_MAX_SOCKS)
		return NULL;

	sx_xlock(&socks_lock);
	sys_sock = socks[sockfd];
	sx_xunlock(&socks_lock);
	return sys_sock;
}

static int
core_itf_sock_close(struct sock_spec *sock_spec)
{
	uint32_t sockfd = sock_spec->sockfd;
	sock_t *sys_sock;

	if (sockfd >= TL_MAX_SOCKS)
		return -EINVAL;

	sx_xlock(&socks_lock);
	sys_sock = socks[sockfd];
	socks[sockfd] = NULL;
	sx_xunlock(&socks_lock);
	if (!sys_sock)
		return -EINVAL;
	sys_sock_close(sys_sock, sock_spec->linger);
	sys_sock_free(sys_sock);
	return 0;
}

static int
__core_itf_sock_create(struct sock_spec *sock_spec, int local)
{
	sock_t *sys_sock;
	int i;

	sys_sock = __sys_sock_create(NULL, local ? AF_LOCAL : AF_INET);
	if (!sys_sock)
		return -ENOMEM;
	sys_sock->usock = local;

	sx_xlock(&socks_lock);
	for (i = 1; i < TL_MAX_SOCKS; i++) {
		if (socks[i])
			continue;
		socks[i] = sys_sock;
		sock_spec->sockfd = i;
		break;
	}
	sx_xunlock(&socks_lock);

	if (i == TL_MAX_SOCKS) {
		sys_sock_close(sys_sock, 0);
		return -1;
	}
	return 0;
}

static int
core_itf_sock_create(struct sock_spec *sock_spec)
{
	return __core_itf_sock_create(sock_spec, 0);
}

static int
core_itf_sock_create_local(struct sock_spec *sock_spec)
{
	return __core_itf_sock_create(sock_spec, 1);
}

static int
core_itf_sock_connect(struct sock_spec *sock_spec)
{
	sock_t *sys_sock;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;

	return sys_sock_connect(sys_sock, sock_spec->ipaddr, sock_spec->local_ipaddr, sock_spec->port);
}

static int
core_itf_sock_shutdown(struct sock_spec *sock_spec)
{
	sock_t *sys_sock;
	int retval;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;

	retval = sys_sock_shutdown(sys_sock);
	atomic_set_bit(SOCK_FLAGS_HAS_SHUTDOWN, &sys_sock->flags);
	chan_wakeup(&sys_sock->sock_wait);
	return retval;
}

static int
core_itf_sock_rw_page(sock_t *sys_sock, struct sock_data_spec *sock_spec, int write)
{
	pagestruct_t *page;
	struct mmap_entry *ret = NULL;
	int retval;

	page = mmap_lookup_page(sock_spec->addr, &ret);
	if (!page)
		return -EIO;

	if (write)
		retval = sys_sock_write_page(sys_sock, page, sock_spec->offset, sock_spec->len);
	else
		retval = sys_sock_read(sys_sock, vm_pg_address(page)+sock_spec->offset, sock_spec->len);
	if (retval < 0)
		return retval;
	sock_spec->done = retval;
	return 0;
}

static int
core_itf_sock_rw(struct sock_data_spec *sock_spec, int write)
{
	sock_t *sys_sock;
	uint8_t *ptr;
	int retval, error;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;

	if (sock_spec->page)
		return core_itf_sock_rw_page(sys_sock, sock_spec, write);

	ptr = kmalloc(sock_spec->len, GFP_KERNEL);
	if (!ptr)
		return -ENOMEM;

	if (write) {
		retval = copy_from_user(ptr, (uint8_t *)sock_spec->addr, sock_spec->len);
		if (retval != 0) {
			kfree(ptr);
			return retval;
		}
		retval = sys_sock_write(sys_sock, ptr, sock_spec->len);
	} else
		retval = sys_sock_read(sys_sock, ptr, sock_spec->len);
	if (retval < 0) {
		kfree(ptr);
		return retval;
	}
	sock_spec->done = retval;
	if (!write) {
		error = copy_to_user((uint8_t *)sock_spec->addr, ptr, retval);
		if (error != 0) {
			kfree(ptr);
			return error;
		}
	}
	kfree(ptr);
	return 0;
}

static int
core_itf_sock_connect_local(struct sock_local_spec *sock_spec)
{
	sock_t *sys_sock;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;

	return sys_sock_connect_local(sys_sock, sock_spec->path);
}

static int
core_itf_sock_poll(struct sock_spec *sock_spec)
{
	sock_t *sys_sock;
	unsigned long timeout = sock_spec->ipaddr;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;

	if (sock_spec->set)
		sock_spec->status = wait_on_chan_timeout(sys_sock->sock_wait, sys_sock_has_write_space(sys_sock) || atomic_test_bit(SOCK_FLAGS_HAS_SHUTDOWN, &sys_sock->flags) || sock_state_error(sys_sock), timeout);
	else
		sock_spec->status = wait_on_chan_timeout(sys_sock->sock_wait, sys_sock_has_read_data(sys_sock) || atomic_test_bit(SOCK_FLAGS_HAS_SHUTDOWN, &sys_sock->flags) || sock_state_error(sys_sock), timeout);
	if (atomic_test_bit(SOCK_FLAGS_IS_CLOSED, &sys_sock->flags) || atomic_test_bit(SOCK_FLAGS_HAS_SHUTDOWN, &sys_sock->flags))
		return -ENOTCONN;
	else
		return 0;
}

static int
core_itf_sock_bind(struct sock_spec *sock_spec)
{
	sock_t *sys_sock;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;

	return sys_sock_bind(sys_sock, sock_spec->ipaddr, sock_spec->port);
}

static int
core_itf_sock_has_write_space(struct sock_spec *sock_spec)
{
	sock_t *sys_sock;
	int retval;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;
	retval = sys_sock_has_write_space(sys_sock);
	if (retval < 0)
		return retval;
	sock_spec->status = retval;
	return 0;
}

static int
core_itf_sock_has_read_data(struct sock_spec *sock_spec)
{
	sock_t *sys_sock;
	int retval;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;
	retval = sys_sock_has_read_data(sys_sock);
	if (retval < 0)
		return retval;
	sock_spec->status = retval;
	return 0;
}

static int
core_itf_sock_is_connected(struct sock_spec *sock_spec)
{
	sock_t *sys_sock;
	int retval;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;
	retval = sys_sock_is_connected(sys_sock);
	if (retval < 0)
		return retval;
	sock_spec->status = retval;
	return 0;
}

static int
core_itf_sock_nopush(struct sock_spec *sock_spec)
{
	sock_t *sys_sock;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;
	sys_sock_nopush(sys_sock, sock_spec->set);
	return 0;
}

static int
core_itf_sock_accept(struct sock_spec *sock_spec)
{
	sock_t *new_syssock;
	sock_t *sys_sock;
	int error, i;

	sys_sock = get_sys_sock(sock_spec->sockfd);
	if (!sys_sock)
		return -EINVAL;
	new_syssock = sys_sock_accept(sys_sock, NULL, &error, &sock_spec->ipaddr);
	if (!new_syssock)
		return error;

	sx_xlock(&socks_lock);
	for (i = 1; i < TL_MAX_SOCKS; i++) {
		if (socks[i])
			continue;
		socks[i] = new_syssock;
		sock_spec->acceptfd = i;
		break;
	}
	sx_xunlock(&socks_lock);

	if (i == TL_MAX_SOCKS) {
		sys_sock_close(new_syssock, 0);
		sys_sock_free(new_syssock);
		return -ENOMEM;
	}
	return 0;
}

static int
core_itf_bdev_sync(struct bdev_sync *sync)
{
	iodev_t *iodev;

	if (sync->kfd < 0 || sync->kfd >= TL_MAX_KFDS)
		return -EINVAL;
	iodev = bdevs[sync->kfd];
	if (!iodev)
		return -EINVAL;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
	return blkdev_issue_flush(file_bdev(iodev));
#elif (defined(CONFIG_SUSE_KERNEL) && LINUX_VERSION_CODE >= KERNEL_VERSION(5, 3, 18) && LINUX_VERSION_CODE <= KERNEL_VERSION(5, 4, 0))
	{
		struct request_queue *q;
		struct bio *bio;
		int retval;

		if (!iodev->bd_disk)
			return -EINVAL;

		q = bdev_get_queue(iodev);
		if (!q)
			return -EINVAL;

		bio = bio_alloc(GFP_NOIO, 0);
		if (!bio)
			return -ENOMEM;
		bio_set_dev(bio, iodev);
		bio_set_op_attrs(bio, REQ_OP_FLUSH, REQ_PREFLUSH);
		retval = submit_bio_wait(bio);
		bio_put(bio);
		return retval;
	}
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(5,12,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8 && RHEL_MINOR >= 6)
	return blkdev_issue_flush(iodev);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8 && RHEL_MINOR >= 4)
	return blkdev_issue_flush(iodev, GFP_NOIO);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0))
	return blkdev_issue_flush(iodev, GFP_NOIO, NULL);
#else
	return blkdev_issue_flush(iodev, NULL);
#endif
}

static int
core_itf_bdev_validate(struct bdev_spec *spec)
{
	iodev_t *b_dev;
	int retval;

	if (spec->kfd >= TL_MAX_KFDS)
		return -1;

	b_dev = bdevs[spec->kfd];
	if (!b_dev)
		return -1;

	retval = validate_block_device(spec->name, b_dev);
	spec->error = retval;
	return 0;
}

static int
core_itf_tdev_close(struct tdev_spec *spec)
{
	struct scsi_device *device;

	DEBUG_INFO_NEW("kfd %lu\n", spec->kfd);
	if (spec->kfd >= TL_MAX_KFDS)
		return -1;

	sx_xlock(&tdevs_lock);
	device = tdevs[spec->kfd];
	tdevs[spec->kfd] = NULL;
	sx_xunlock(&tdevs_lock);
	if (!device) {
		DEBUG_WARN_NEW("Cannot find device at %lu\n", spec->kfd);
		return -1;
	}
	scsi_device_put(device);
	DEBUG_INFO_NEW("closed kfd %lu\n", spec->kfd);
	return 0;
}

static int
core_itf_bdev_close(struct bdev_spec *spec)
{
	iodev_t *b_dev;
	if (spec->kfd >= TL_MAX_KFDS)
		return -1;

	sx_xlock(&bdevs_lock);
	b_dev = bdevs[spec->kfd];
	bdevs[spec->kfd] = NULL;
	sx_xunlock(&bdevs_lock);
	if (!b_dev)
		return -1;
	core_itf_iodev_close(b_dev);
	return 0;
}

#if (defined(RHEL_MAJOR) && RHEL_MAJOR >= 7) || (LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
static int
core_itf_file_truncate(struct file_spec *spec)
{
	file_t *fp = (file_t *)spec->kfd;
	uint64_t length = spec->length;
	struct path *path = &fp->f_path;
	struct inode *ino = fp->f_path.dentry->d_inode;
	struct iattr iattr;
	uint64_t cur_length;
	int retval;

	cur_length = i_size_read(ino);
	if (cur_length <= length)
		return 0;
	memset(&iattr, 0, sizeof(iattr));
	iattr.ia_valid = ATTR_SIZE;
	iattr.ia_size = length;
	inode_lock(ino);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,5,0)) || (defined(CONFIG_SUSE_KERNEL) && LINUX_VERSION_CODE >= KERNEL_VERSION(6, 4, 0))
	retval = notify_change(&nop_mnt_idmap, path->dentry, &iattr, NULL);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(5,12,0))
	retval = notify_change(&init_user_ns, path->dentry, &iattr, NULL);
#else
	retval = notify_change(path->dentry, &iattr, NULL);
#endif
	inode_unlock(ino);
	return retval;
}
#else
static int
core_itf_file_truncate(struct file_spec *spec)
{
	return -EIO;
}
#endif

static int
core_itf_file_sync(struct file_spec *spec)
{
	file_t *fp = (file_t *)spec->kfd;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0))
	return vfs_fsync(fp, 0);
#else
	return vfs_fsync(fp, fp->f_path.dentry, 0);
#endif
}

static int
core_itf_file_close(struct file_spec *fspec)
{
	filp_close((file_t *)fspec->kfd, NULL);
	return 0;
}

static int
core_itf_file_open(struct file_spec *spec)
{
	file_t *fp;
	int flags = O_RDWR | O_LARGEFILE;
	int mode = 0;

	if (spec->isnew) {
		flags |= O_CREAT | O_TRUNC;
		mode = 0600;
	}

	fp = filp_open(spec->name, flags, mode);
	if (IS_ERR(fp))
		return -EIO;

	spec->kfd = (unsigned long)fp;
	return 0;
}

static int
core_itf_tdev_open(struct tdev_spec *spec)
{
	struct Scsi_Host *hostptr;
	struct scsi_device *device;
	int i;

	DEBUG_INFO_NEW("Open device %s host %u:%u:%u:%llu\n", spec->name, spec->host, spec->channel, spec->target, (unsigned long long)spec->lun);
	hostptr = scsi_host_lookup(spec->host);
	if (IS_ERR(hostptr)) {
		DEBUG_WARN_NEW("Unable to locate host at %u\n", spec->host);
		return -1;
	}

	device = scsi_device_lookup(hostptr, spec->channel, spec->target, spec->lun);
	scsi_host_put(hostptr);
	if (!device) {
		DEBUG_WARN_NEW("Unable to locate device at %u %u %u %llu\n", spec->host, spec->channel, spec->target, (unsigned long long)spec->lun);
		return -1;
	}

	sx_xlock(&tdevs_lock);
	for (i = 0; i < TL_MAX_KFDS; i++) {
		if (tdevs[i])
			continue;
		tdevs[i] = device;
		spec->kfd = i;
		break;
	}
	sx_xunlock(&tdevs_lock);

	if (i == TL_MAX_KFDS) {
		scsi_device_put(device);
		return -1;
	}
	DEBUG_INFO_NEW("device %s kfd %lu\n", spec->name, spec->kfd);
	return 0;
}
 
static int
core_itf_bdev_open(struct bdev_spec *spec)
{
	iodev_t *b_dev;
	int i;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
	b_dev = bdev_file_open_by_path(spec->name, BLK_OPEN_READ | BLK_OPEN_WRITE, THIS_MODULE, NULL);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(6,5,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR == 4) || (defined(CONFIG_SUSE_KERNEL) && LINUX_VERSION_CODE >= KERNEL_VERSION(6, 4, 0))
	b_dev = blkdev_get_by_path(spec->name, FMODE_READ | FMODE_WRITE, THIS_MODULE, NULL);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0))
	b_dev = blkdev_get_by_path(spec->name, FMODE_READ | FMODE_WRITE, THIS_MODULE);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29))
	b_dev = open_bdev_exclusive(spec->name, FMODE_READ | FMODE_WRITE, THIS_MODULE);
#else
	b_dev = open_bdev_excl(spec->name, 0, THIS_MODULE);
#endif
	if (unlikely(IS_ERR(b_dev))) {
		DEBUG_WARN_NEW("Unable to open dev %s err is %ld\n", spec->name, PTR_ERR(b_dev));
		return -1;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
	spec->sector_size = bdev_logical_block_size(file_bdev(b_dev));
	spec->size = file_bdev(b_dev)->bd_inode->i_size;
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
	spec->sector_size = bdev_hardsect_size(b_dev);
	spec->size = b_dev->bd_inode->i_size;
#else
	spec->sector_size = bdev_logical_block_size(b_dev);
	spec->size = b_dev->bd_inode->i_size;
#endif

	sx_xlock(&bdevs_lock);
	for (i = 0; i < TL_MAX_KFDS; i++) {
		if (bdevs[i])
			continue;
		bdevs[i] = b_dev;
		spec->kfd = i;
		break;
	}
	sx_xunlock(&bdevs_lock);

	if (i == TL_MAX_KFDS) {
		core_itf_iodev_close(b_dev);
		return -1;
	}
	return 0;
}

struct bio_priv {
	uint64_t b_start;
	uint32_t kfd;
	unsigned long bio_addr;
	atomic_t pending;
	atomic_t error;
	TAILQ_ENTRY(bio_priv) b_entry;
};
TAILQ_HEAD(bio_priv_list, bio_priv);

struct bio_priv_list bio_priv_list = TAILQ_HEAD_INITIALIZER(bio_priv_list);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
int static bio_end_bio(bio_t *bio, uint32_t bytes_done, int err)
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(4,4,0))
void static bio_end_bio(bio_t *bio, int err)
#else
void static bio_end_bio(bio_t *bio)
#endif
{
	struct bio_priv *bpriv = (struct bio_priv *)(bio->bi_private);
#if ((defined(CONFIG_SUSE_KERNEL) && (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))) || (LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)))
	int err = bio->bi_status;
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,0))
	int err = bio->bi_error;
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
	if (bio->bi_size)
		return 1;
#endif

	if (err) {
		if (err < 0)
			err = -(err);
		DEBUG_WARN_NEW("bio err for %llu %u with status %d\n", (unsigned long long)bpriv->b_start, bpriv->kfd, err);
		atomic_set(&bpriv->error, err);
	}

	if (atomic_dec_and_test(&bpriv->pending)) {
		unsigned long flags;
		chan_lock_intr(&bio_chan, &flags);
		TAILQ_INSERT_TAIL(&bio_priv_list, bpriv, b_entry);
		chan_wakeup(&bio_chan);
		chan_unlock_intr(&bio_chan, &flags);
	}

	bio_put(bio);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24))
	return 0;
#endif
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 0, 0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 9) || (defined(CONFIG_SUSE_KERNEL) && CONFIG_SUSE_VERSION >= 15 && CONFIG_SUSE_PATCHLEVEL > 4)
static struct bio *
bio_setup(struct bio_priv *bpriv, iodev_t *iodev, uint64_t bi_sector, int rw, int file, int bio_vec_count)
{
	struct bio *bio;
	uint32_t opf = 0;

	switch (rw) {
	case QS_IO_READ:
		opf = REQ_OP_READ;
		break;
	case QS_IO_WRITE:
		opf = REQ_OP_WRITE;
		break;
	case QS_IO_SYNC:
		opf = (REQ_OP_WRITE | REQ_SYNC | REQ_FUA);
		break;
	case QS_IO_SYNC_FLUSH:
		opf = (REQ_OP_WRITE | REQ_SYNC | REQ_PREFLUSH | REQ_FUA);
		break;
	default:
		DEBUG_BUG_ON(1);
		return NULL;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
	bio = bio_alloc(file_bdev(iodev), bio_vec_count, opf, M_NOWAIT);
#else
	bio = bio_alloc(iodev, bio_vec_count, opf, M_NOWAIT);
#endif
	if (!bio)
		return NULL;
	bio->bi_iter.bi_sector = bi_sector;
	bio->bi_end_io = bio_end_bio;
	bio->bi_private = bpriv;
	atomic_inc(&bpriv->pending);
	return bio;
}
#else
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 8, 0)) || (defined(CONFIG_SUSE_KERNEL) && LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 73))
static void
bio_set_command(bio_t *bio, int cmd)
{
	switch (cmd) {
	case QS_IO_READ:
		bio_set_op_attrs(bio, REQ_OP_READ, 0);
		break;
	case QS_IO_WRITE:
		bio_set_op_attrs(bio, REQ_OP_WRITE, 0);
		break;
	case QS_IO_SYNC:
		bio_set_op_attrs(bio, REQ_OP_WRITE, WRITE_FUA);
		break;
	case QS_IO_SYNC_FLUSH:
		bio_set_op_attrs(bio, REQ_OP_WRITE, WRITE_FLUSH_FUA);
		break;
	default:
		DEBUG_BUG_ON(1);
	}
}
#else
static void
bio_set_command(bio_t *bio, int cmd)
{
	switch (cmd) {
	case QS_IO_READ:
		bio->bi_rw = READ;
		break;
	case QS_IO_WRITE:
		bio->bi_rw = WRITE;
		break;
	case QS_IO_SYNC:
#if defined(WRITE_FUA)
		bio->bi_rw = WRITE_FUA;
#elif defined(WRITE_BARRIER)
		bio->bi_rw = WRITE_BARRIER;
#else
#error "Cannot handle sync cmds"
#endif
		break;
	case QS_IO_SYNC_FLUSH:
#if defined(WRITE_FLUSH_FUA)
		bio->bi_rw = WRITE_FLUSH_FUA;
#elif defined(WRITE_BARRIER)
		bio->bi_rw = WRITE_BARRIER;
#else
#error "Cannot handle sync cmds"
#endif
		break;
	default:
		DEBUG_BUG_ON(1);
	}
}
#endif
static struct bio *
bio_setup(struct bio_priv *bpriv, iodev_t *iodev, uint64_t bi_sector, int rw, int file, int bio_vec_count)
{
	struct bio *bio;

	bio = bio_alloc(M_NOWAIT, bio_vec_count);
	if (!bio)
		return NULL;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,14,0))
	bio->bi_iter.bi_sector = bi_sector;
#else
	bio->bi_sector = bi_sector;
#endif
#if ((defined(CONFIG_SUSE_KERNEL) && (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))) || (LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)))
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,14,0)
	bio_set_dev(bio, iodev);
#else
	bio->bi_disk = iodev->bd_disk;
	bio->bi_partno = iodev->bd_partno;
#endif
#else
	bio->bi_bdev = iodev;
#endif
	bio->bi_end_io = bio_end_bio;
	bio->bi_private = bpriv;
	bio_set_command(bio, rw);
	atomic_inc(&bpriv->pending);
	return bio;
}
#endif

extern void * vm_pg_map(pagestruct_t **pp, int pg_count);
extern void vm_pg_unmap(void *maddr, int pg_count);

void
relay_ctio_free_data(struct qsio_scsiio *ctio)
{
}

extern int __relay_cmnd(void *mdevice, struct qsio_scsiio *ctio, int *ret_resid, int usr);
static int
core_itf_relay_cmnd(struct ctio_tape_req *tape_req, struct pgdata_spec *spec, uint8_t *data)
{
	struct scsi_device *device;
	struct qsio_scsiio *ctio;
	struct pgdata *pglist, *dest_pgdata;
	struct pgdata_spec *src_pgdata;
	struct mmap_entry *ret = NULL;
	int i, retval;

	if (tape_req->kfd >= TL_MAX_KFDS) {
		DEBUG_WARN_NEW("Invalid kfd %lu\n", tape_req->kfd);
		return -1;
	}

	device = tdevs[tape_req->kfd];
	if (!device) {
		DEBUG_WARN_NEW("Cannot find device at %lu\n", tape_req->kfd);
		return -1;
	}

	ctio = ctio_new(0);
	if (!ctio)
		return -ENOMEM;
	memcpy(ctio->cdb, tape_req->cdb, 16);
	if (tape_req->pglist_cnt) {
		pglist = pgdata_allocate_pglist(tape_req->pglist_cnt, 0);
		if (!pglist) {
			__uma_zfree(ctio_cache, ctio);
			return -ENOMEM;
		}
		src_pgdata = &spec[0];
		dest_pgdata = &pglist[0];
		for (i = 0; i < tape_req->pglist_cnt; i++, src_pgdata++, dest_pgdata++) {
			dest_pgdata->page = mmap_lookup_page(src_pgdata->addr, &ret);
			if (!dest_pgdata->page) {
				__qs_free(pglist, M_PGLIST);
				__uma_zfree(ctio_cache, ctio);
				return -EINVAL;
			}
			dest_pgdata->pg_len = src_pgdata->pg_len;
		}
		ctio->data_ptr = (uint8_t *)pglist;
		ctio->pglist_cnt = tape_req->pglist_cnt;
	}
	else if (tape_req->dxfer_len)
		ctio->data_ptr = data;

	ctio->dxfer_len = tape_req->dxfer_len;
	retval = __relay_cmnd(device, ctio, &tape_req->resid, 1);
	tape_req->scsi_status = ctio->scsi_status;
	if (tape_req->pglist_cnt) 
		__qs_free((struct pgdata *)ctio->data_ptr, M_PGLIST);
	if (retval == 0 && ctio->dxfer_len && tape_req->dir == DATA_DIR_FROM_TAPE) {
		if (ctio->dxfer_len > tape_req->dxfer_len)
			DEBUG_WARN_NEW("Invalid ctio dxfer len %d for %x tape req dxfer len %d\n", ctio->dxfer_len, ctio->cdb[0], tape_req->dxfer_len);
		else
			tape_req->dxfer_len = ctio->dxfer_len;
		if (!tape_req->pglist_cnt) {
			retval = copyout(data, tape_req->data, tape_req->dxfer_len);
			if (retval != 0)
				DEBUG_WARN_NEW("Cannot copy out %d bytes\n", tape_req->dxfer_len);
		}
	}

	if (ctio->sense_len) {
		memcpy(tape_req->sense, ctio->sense_data, ctio->sense_len);
		tape_req->sense_len = ctio->sense_len;
	}

	__uma_zfree(ctio_cache, ctio);

	return retval;
}
 
static int
core_itf_file_req(struct bio_spec *spec, struct pgdata_spec *pglist)
{
	file_t *fp = (file_t *)(spec->kfd);
	struct pgdata_spec *pgdata_spec;
	pagestruct_t **pages;
	struct mmap_entry *ret = NULL;
	uint8_t *ptr;
	struct bio_priv *bpriv;
	unsigned long flags;
	mm_segment_t saved_fs;
	loff_t fpos;
	int retval, err = 0, i;

	bpriv = kzalloc(sizeof(*bpriv), GFP_KERNEL);
	if (!bpriv)
		return -ENOMEM;

	fpos = (spec->b_start << 9);
	pages = __qs_malloc(sizeof(pagestruct_t *) * spec->page_count, M_FILE, M_NOWAIT);
	if (!pages) {
		DEBUG_WARN_NEW("Cannot allocate for cpage count %d\n", spec->page_count);
		kfree(bpriv);
		return -ENOMEM;
	}

	for (i = 0; i < spec->page_count; i++) {
		pgdata_spec = &pglist[i];
		pages[i] = mmap_lookup_page(pgdata_spec->addr, &ret);
		if (!pages[i]) {
			free(pages, M_FILE);
			kfree(bpriv);
			return -EIO;
		}
	}

	ptr = vm_pg_map(pages, spec->page_count);
	if (!ptr) {
		DEBUG_WARN_NEW("Cannot vm page map for page_count %d\n", spec->page_count);
		free(pages, M_FILE);
		kfree(bpriv);
		return -EIO;
	}

	saved_fs = get_fs();
	set_fs(KERNEL_DS);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8)
	if (spec->rw == QS_IO_READ)
		retval = kernel_read(fp, ptr, spec->length, &fpos);
	else
		retval = kernel_write(fp, ptr, spec->length, &fpos);
#else
	if (spec->rw == QS_IO_READ)
		retval = vfs_read(fp, ptr, spec->length, &fpos);
	else
		retval = vfs_write(fp, ptr, spec->length, &fpos);
#endif
	set_fs(saved_fs);
	vm_pg_unmap(ptr, spec->page_count);
	free(pages, M_FILE);
	if (retval != spec->length)
		err = -EIO;
	if (err) {
		kfree(bpriv);
		return err;
	}

	bpriv->bio_addr = spec->bio_addr;
	chan_lock_intr(&bio_chan, &flags);
	TAILQ_INSERT_TAIL(&bio_priv_list, bpriv, b_entry);
	chan_wakeup(&bio_chan);
	chan_unlock_intr(&bio_chan, &flags);
	return 0;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,0,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 9) || (defined(CONFIG_SUSE_KERNEL) && LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 21))
static int
bdev_unmap_support(iodev_t *iodev)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
	if (bdev_max_discard_sectors(file_bdev(iodev)))
#else
	if (bdev_max_discard_sectors(iodev))
#endif
		return 0;
	else
		return -EINVAL;
}
#else
static int
bdev_unmap_support(iodev_t *iodev)
{
	struct request_queue *q = bdev_get_queue(iodev);

	if (blk_queue_discard(q))
		return 0;
	else
		return -EINVAL;
}
#endif

static int
core_itf_bdev_unmap_support(struct bio_spec *spec)
{
	iodev_t *iodev;

	if (spec->kfd >= TL_MAX_KFDS)
		return -EINVAL;

	iodev = bdevs[spec->kfd];
	if (!iodev)
		return -EINVAL;


	return bdev_unmap_support(iodev);
}

static int
core_itf_bio_unmap(struct bio_spec *spec)
{
	iodev_t *iodev;
	int retval;

	if (spec->kfd >= TL_MAX_KFDS)
		return -EINVAL;

	iodev = bdevs[spec->kfd];
	if (!iodev)
		return -EINVAL;

	DEBUG_INFO_NEW("spec b_start %llu length %u\n", (unsigned long long)spec->b_start, spec->length);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
	retval = blkdev_issue_discard(file_bdev(iodev), spec->b_start, spec->length, GFP_NOIO);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(6,0,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9) || (defined(CONFIG_SUSE_KERNEL) && LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 21))
	retval = blkdev_issue_discard(iodev, spec->b_start, spec->length, GFP_NOIO);
#else
	retval = blkdev_issue_discard(iodev, spec->b_start, spec->length, GFP_NOIO, 0);
#endif

	return retval;
}

static int
core_itf_bio_req(struct bio_spec *spec, struct pgdata_spec *pglist)
{
	iodev_t *iodev;
	struct bio_list bio_list;
	struct bio *bio;
	int todo = spec->page_count;
	int bio_vec_count = spec->page_count > 128 ? 128 : spec->page_count;
	uint64_t b_start = spec->b_start;
	struct pgdata_spec *pgdata_spec;
	struct mmap_entry *ret = NULL;
	pagestruct_t *page;
	struct bio_priv *bpriv;
	int idx = 0, bio_length = 0, retval;

	if (spec->file)
		return core_itf_file_req(spec, pglist);

	bpriv = kzalloc(sizeof(*bpriv), GFP_KERNEL);
	if (!bpriv)
		return -ENOMEM;

	bio_list_init(&bio_list);
	atomic_set(&bpriv->pending, 0);
	atomic_set(&bpriv->error, 0);

	if (spec->kfd >= TL_MAX_KFDS) {
		kfree(bpriv);
		return -EINVAL;
	}

	iodev = bdevs[spec->kfd];
	if (!iodev) {
		kfree(bpriv);
		return -EINVAL;
	}

	bio = bio_setup(bpriv, iodev,  b_start, spec->rw, spec->file, bio_vec_count);
	if (!bio) {
		kfree(bpriv);
		return -ENOMEM;
	}

	bpriv->bio_addr = spec->bio_addr;
	bpriv->b_start = b_start;
	bpriv->kfd = spec->kfd;
	bio_list_add(&bio_list, bio);

	idx = 0;
	while (todo) {
		pgdata_spec = &pglist[idx];
		page = mmap_lookup_page(pgdata_spec->addr, &ret);
		if (!page)
			goto err;
retry:
		retval = bio_add_page(bio, page, pgdata_spec->pg_len, pgdata_spec->pg_offset);
		if (retval != pgdata_spec->pg_len) {
			if (!bio_length)
				goto err;
			b_start += (bio_length >> 9);
			bio_length = 0;
			bio = bio_setup(bpriv, iodev,  b_start, spec->rw, spec->file, bio_vec_count);
			if (!bio)
				goto err;
			bio_list_add(&bio_list, bio);
			goto retry;
		}
		bio_length += pgdata_spec->pg_len;
		todo--;
		idx++;
	}

	while ((bio = bio_list_pop(&bio_list)) != NULL) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
		submit_bio(bio);
#else
		generic_make_request(bio);
#endif
	}
	return 0;
err:
	while ((bio = bio_list_pop(&bio_list)) != NULL) {
		bio_put(bio);
	}
	return -EIO;
}

static int
core_itf_ctio_wakeup(struct ctio_wakeup *ctio_wakeup)
{
	struct tdevice *tdevice;
	uint32_t tid = (ctio_wakeup->lun_id - 1);

	if (tid >= MAX_DINFO_DEVICES)
		return -1;

	tdevice = device_list[tid];
	if (!tdevice) {
		DEBUG_INFO_NEW("Cannot find tdevice\n");
		return -1;
	}

	complete(&tdevice->ctio_completion);
	return 0;
}

static int
core_itf_bio_mmap_exit(void)
{
	if (!bdevice)
		return -EINVAL;
	atomic_set(&bdevice->disabled, 1);
	wake_up_interruptible(&bio_msg_wait);
	return 0;
}

static int
core_itf_bio_mmap_setup(struct file *file, struct bio_mmap_setup *bio_setup)
{
	struct mmap_entry *ret = NULL;
	pagestruct_t *page;

	bdevice = kzalloc(sizeof(*bdevice), GFP_KERNEL);
	if (!bdevice)
		return -ENOMEM;
	bdevice->type = TYPE_BDEVICE;

	page = mmap_lookup_page((unsigned long)bio_setup->addr, &ret);
	if (!page)
		return -1;

	mmap_entry_remove((unsigned long)bio_setup->addr);
	file->private_data = bdevice;
	msg_addr = vm_pg_address(page);
	memset(msg_addr, 0, BIO_MSG_SIZE);
	return 0;
}

static int
core_itf_ctio_mmap_setup(struct file *file, struct ctio_mmap_setup *ctio_setup)
{
	struct tdevice *tdevice;
	struct mmap_entry *ret = NULL;
	uint32_t tid = (ctio_setup->lun_id - 1);
	pagestruct_t *page;

	DEBUG_INFO_NEW("tid %u lun id %u\n", tid, ctio_setup->lun_id);
	if (tid >= MAX_DINFO_DEVICES)
		return -1;

	tdevice = device_list[tid];
	if (!tdevice)
		return -1;

	page = mmap_lookup_page((unsigned long)ctio_setup->addr, &ret);
	DEBUG_INFO_NEW("lookup page %p\n", page);
	if (!page)
		return -1;

	mmap_entry_remove((unsigned long)ctio_setup->addr);
	tdevice_get(tdevice);
	file->private_data = tdevice;
	atomic_set(&tdevice->pending_msg, 0);
	tdevice->ctio_page = page;
	DEBUG_INFO_NEW("ctio page setup tdevice %p\n", page);
	return 0;
}

static void
free_wrkmem(struct wrkmem *wrkmem)
{
	free(wrkmem->pglist, M_COMP);
	free(wrkmem->pages, M_COMP);
	free(wrkmem->wrkmem, M_COMP);
	free(wrkmem, M_COMP);
}

static void
put_wrkmem(struct wrkmem *wrkmem)
{
	if (atomic_read(&wrkmem_count) > 256) { /* 256 parallel compression threads should be enough */
		free_wrkmem(wrkmem);
		return;
	}

	spin_lock(&wrkmem_lock);
	TAILQ_INSERT_HEAD(&wrkmem_list, wrkmem, w_list);
	atomic_inc(&wrkmem_count);
	spin_unlock(&wrkmem_lock);
}

static struct wrkmem *
alloc_wrkmem(void)
{
	struct wrkmem *wrkmem;

	wrkmem = __qs_malloc(sizeof(*wrkmem), M_COMP, Q_NOWAIT);
	if (!wrkmem) {
		DEBUG_WARN_NEW("Cannot allocate wrkmem\n");
		return NULL;
	}

	/* 64k based on dict size */
	wrkmem->wrkmem = __qs_malloc(65536, M_COMP, Q_NOWAIT);
	if (!wrkmem->wrkmem) {
		free(wrkmem, M_COMP);
		return NULL;
	}

	/* 64k for a possible 4096 uncomp pages and 4096 comp page address. Max tape block size possible 16777216 */
	wrkmem->pages = __qs_malloc(65536, M_COMP, Q_NOWAIT);
	if (!wrkmem->pages) {
		free(wrkmem->wrkmem, M_COMP);
		free(wrkmem, M_COMP);
		return NULL;
	}

	wrkmem->pglist = __qs_malloc(65536, M_COMP, Q_NOWAIT);
	if (!wrkmem->pglist) {
		free(wrkmem->pages, M_COMP);
		free(wrkmem->wrkmem, M_COMP);
		free(wrkmem, M_COMP);
		return NULL;
	}
	return wrkmem;
}

static struct wrkmem *
get_wrkmem(void)
{
	struct wrkmem *wrkmem;

	spin_lock(&wrkmem_lock);
	wrkmem = TAILQ_FIRST(&wrkmem_list);
	if (wrkmem)
		TAILQ_REMOVE(&wrkmem_list, wrkmem, w_list);
	spin_unlock(&wrkmem_lock);

	if (wrkmem) { 
		atomic_dec(&wrkmem_count);
		return wrkmem;
	}

	return alloc_wrkmem();
}

static int
core_itf_ctio_setup(struct ctio_spec *ctio_spec, struct pgdata_spec *pgdatalist)
{
	struct tdevice *tdevice;
	struct mmap_entry *ret = NULL;
	struct qsio_scsiio *ctio, *old_ctio;
	struct pgdata *pglist, *dest_pgdata;
	struct pgdata_spec *src_pgdata;
	uint32_t tid = (ctio_spec->tid - 1);
	int i;

	if (tid >= MAX_DINFO_DEVICES)
		return -EINVAL;

	tdevice = device_list[tid];
	if (!tdevice || !tdevice->ctio_page)
		return -EINVAL;

	if (tdevice->ctio)
		return -EINVAL;

	ctio = ctio_new(0);
	if (!ctio)
		return -ENOMEM;

	pglist = pgdata_allocate_pglist(ctio_spec->pglist_cnt, 0);
	if (!pglist) { 
		__uma_zfree(ctio_cache, ctio);
		return -ENOMEM;
	}

	src_pgdata = &pgdatalist[0];
	dest_pgdata = &pglist[0];
	for (i = 0; i < ctio_spec->pglist_cnt; i++, src_pgdata++, dest_pgdata++) {
		dest_pgdata->page = mmap_lookup_page(src_pgdata->addr, &ret);
		if (!dest_pgdata->page) {
			__qs_free(pglist, M_PGLIST);
			__uma_zfree(ctio_cache, ctio);
			return -EINVAL;
		}
		dest_pgdata->pg_len = src_pgdata->pg_len;
	}
	ctio->data_ptr = (void *)pglist;
	ctio->dxfer_len = ctio_spec->dxfer_len; 
	ctio->pglist_cnt = ctio_spec->pglist_cnt;
	ctio->tx_id = ctio_spec->tx_id;
	ctio->ccb_h.flags |= QSIO_USR_PAGES;
	ctio->ccb_h.flags |= QSIO_DATA_DIR_OUT;
	memcpy(ctio->cdb, ctio_spec->cdb, 16);
	ctio->ccb_h.tdevice = tdevice;

	spin_lock(&tdevice->ctio_alloc_lock);
	old_ctio = tdevice->ctio;
	if (!old_ctio)
		tdevice->ctio = ctio;
	spin_unlock(&tdevice->ctio_alloc_lock);
	if (old_ctio) {
		__qs_free(pglist, M_PGLIST);
		__uma_zfree(ctio_cache, ctio);
		return -EINVAL;
	}
	return 0;
}

static int
core_itf_comp_uncomp(struct comp_spec *comp_spec, struct page_spec *cpages, struct page_spec *upages, struct wrkmem *wrkmem)
{
	struct page_spec *spec;
	pagestruct_t **cpglist = NULL;
	pagestruct_t **upglist = NULL;
	struct mmap_entry *ret = NULL;
	uint8_t *cptr = NULL, *uptr = NULL;
	int i, error = -ENOMEM, retval;

	cpglist = (pagestruct_t **)wrkmem->pglist;
	upglist = cpglist + comp_spec->cpage_count;

	for (i = 0; i < comp_spec->cpage_count; i++) {
		spec = &cpages[i];
		cpglist[i] = mmap_lookup_page((unsigned long)spec->addr, &ret);
		if (!cpglist[i]) {
			DEBUG_WARN_NEW("Cannot lookup cpage %lu\n", spec->addr);
			error = -EINVAL;
			goto err;
		}
	}

	for (i = 0; i < comp_spec->upage_count; i++) {
		spec = &upages[i];
		upglist[i] = mmap_lookup_page((unsigned long)spec->addr, &ret);
		if (!upglist[i]) {
			DEBUG_WARN_NEW("Cannot lookup upage %lu\n", spec->addr);
			error = -EINVAL;
			goto err;
		}
	}

	cptr = vm_pg_map(cpglist, comp_spec->cpage_count);
	if (!cptr) {
		DEBUG_WARN_NEW("Cannot vm page map for cpage_count %d\n", comp_spec->cpage_count);
		goto err;
	}

	uptr = vm_pg_map(upglist, comp_spec->upage_count);
	if (!uptr) {
		DEBUG_WARN_NEW("Cannot vm page map for upage_count %d\n", comp_spec->upage_count);
		goto err;
	}

	error = 0;
	if (comp_spec->compress) {
		retval = qs_deflate_block(uptr, comp_spec->uncomp_size, cptr, &comp_spec->comp_size, wrkmem->wrkmem, COMP_ALG_LZ4);
		if (retval != 0)
			comp_spec->comp_size = 0;
	}
	else {
		retval = qs_inflate_block(cptr, comp_spec->comp_size, uptr, comp_spec->uncomp_size);
		if (retval != 0)
			error = -EIO;
	}
err:
	if (cptr)
		vm_pg_unmap(cptr, comp_spec->cpage_count);
	if (uptr)
		vm_pg_unmap(uptr, comp_spec->upage_count);
	return error;
}

static int
ldev_qload_done(void)
{
	struct qs_interface_cbs *cbs;

	sx_xlock(&cbs_lock);
	atomic_set(&qload_done, 1);
	LIST_FOREACH(cbs, &cbs_list, i_list) {
		cbs->qload_done = 1;
	}
	sx_xunlock(&cbs_lock);
	return 0;
}

union ioctl_spec {
	struct comp_spec comp_spec;
	struct mdaemon_info mdaemon_info;
	struct ctio_wakeup ctio_wakeup;
	struct ldev_info ldev_info;
	struct bio_spec bio_spec;
	struct ctio_spec ctio_spec;
	struct bdev_sync bdev_sync;
	struct ctio_mmap_setup ctio_setup;
	struct bio_mmap_setup bio_setup;
	struct ctio_tape_req tape_req;
	struct file_spec *file_spec;
	struct sock_spec sock_spec;
	struct sock_local_spec sock_local_spec;
	struct sock_data_spec sock_data_spec;
	struct uma_spec uma_spec;
};

static void 
ioctl_send_ccb(void *ccb_void)
{
	struct qsio_scsiio *ctio = ccb_void;
	struct completion *compl = ctio->ccb_h.priv.cpriv.completion;

	complete(compl);
}

static struct qsio_scsiio *
ctio_alloc(uint32_t device_tid)
{
	struct qsio_scsiio *ctio;
	struct tdevice *tdevice;

	if (!device_tid || device_tid >= MAX_DINFO_DEVICES)
		return NULL;

	tdevice = device_list[device_tid -1];
	if (!tdevice)
		return NULL;

	ctio = ctio_new(0);
	if (!ctio)
		return NULL;

	ctio->init_int = TARGET_INT_LOCAL;
	ctio->i_prt[0] = 14;
	ctio->t_prt[0] = 1;
	ctio->r_prt = LDEV_RPORT_START; 
	ctio->ccb_h.queue_fn = ioctl_send_ccb;
	ctio->task_attr = MSG_SIMPLE_TASK; 
	ctio->ccb_h.flags = QSIO_DIR_OUT;
	ctio->ccb_h.tdevice = tdevice;
	return ctio;
}

extern int ctio_copy_tape_data(struct qsio_scsiio *ctio, uint8_t *buffer, int read_from_user, uint32_t *csum);
static void
ctio_queue(struct qsio_scsiio *ctio)
{
	DECLARE_COMPLETION_ONSTACK(compl);

	ctio->ccb_h.priv.cpriv.completion = &compl;
	device_queue_ctio(ctio->ccb_h.tdevice, ctio);
	wait_for_completion(&compl);
}


#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0))
static long coreitf_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#else
static int coreitf_ioctl(vnode_t *i, struct file *file, uint32_t cmd, unsigned long arg)
#endif
{
	void __user *userp = (void __user *)arg;
	uint8_t *data;
	int err = 0;
	int retval = -EINVAL;
	struct bdev_info *bdev_info;
	union ioctl_spec req;
	struct mdaemon_info *mdaemon_info;
	struct node_config *node_config;
	struct group_conf *group_conf;
	struct mirror_config *mirror_config;
	struct clone_config *clone_config;
	struct vdeviceinfo *deviceinfo;
	struct vcartridge *vcartridge;
	struct fc_rule_config fc_rule_config;
	struct bdev_spec *bdev_spec;
	struct tdev_spec *tdev_spec;
	struct bio_spec *bio_spec;
	struct ctio_spec *ctio_spec;
	struct comp_spec *comp_spec;
	struct ctio_tape_req *tape_req;
	struct page_spec *cpages, *upages;
	struct pgdata_spec *pglist;
	struct map_slot *mslot;
	struct wrkmem *wrkmem;
	struct ieinfo_common *ieinfo;
	struct libctl *libctl;
	struct vdevicedump *vdevicedump;
	struct ctio_mmap_setup *ctio_setup;
	struct bio_mmap_setup *bio_setup;
	struct ctio_wakeup *ctio_wakeup;
	struct qsio_scsiio *ctio;
	struct tapebuf *tapebuf;
	struct tapecmd *tapecmd;
	struct sock_spec *sock_spec;
	struct sock_local_spec *sock_local_spec;
	struct sock_data_spec *sock_data_spec;
	struct uma_spec *uma_spec;
	uint8_t *buffer;

	/* Check the capabilities of the user */
	if (!capable(CAP_SYS_ADMIN))
		return (-EPERM);

	if (_IOC_TYPE (cmd) != TL_MAGIC)
		return -ENOTTY;

	err = 0;
	if (_IOC_DIR (cmd) & _IOC_READ) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8 && RHEL_MINOR >= 1)
		err = !access_ok (userp, _IOC_SIZE (cmd));
#else
		err = !access_ok (VERIFY_READ, userp, _IOC_SIZE (cmd));
#endif
	}

	if (err)
		return -EPERM;

	err = 0;
	if (_IOC_DIR (cmd) & _IOC_WRITE) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8 && RHEL_MINOR >= 1)
		err = !access_ok (userp, _IOC_SIZE (cmd));
#else
		err = !access_ok (VERIFY_WRITE, userp, _IOC_SIZE (cmd));
#endif
	}

	if (err)
		return -EPERM;

	switch (cmd) {
	case TLTARGIOCBIOMSGMMAPSETUP:
		bio_setup = &req.bio_setup;
		if ((retval = copyin(userp, bio_setup, sizeof(*bio_setup))) != 0)
			break;

		retval = core_itf_bio_mmap_setup(file, bio_setup);
		if (retval == 0)
			retval = copyout(bio_setup, userp, sizeof(*bio_setup));
		else
			err = copyout(bio_setup, userp, sizeof(*bio_setup));
		break;
	case TLTARGIOCBIOMSGMMAPEXIT:
		retval = core_itf_bio_mmap_exit();
		break;
	case TLTARGIOCCTIOMMAPSETUP:
		ctio_setup = &req.ctio_setup;
		if ((retval = copyin(userp, ctio_setup, sizeof(*ctio_setup))) != 0)
			break;

		retval = core_itf_ctio_mmap_setup(file, ctio_setup);
		if (retval == 0)
			retval = copyout(ctio_setup, userp, sizeof(*ctio_setup));
		else
			err = copyout(ctio_setup, userp, sizeof(*ctio_setup));
		break;
	case TLTARGIOCCTIOWAKEUP:
		ctio_wakeup = &req.ctio_wakeup;
		if ((retval = copyin(userp, ctio_wakeup, sizeof(*ctio_wakeup))) != 0)
			break;
		retval = core_itf_ctio_wakeup(ctio_wakeup);
		break;
	case TLTARGIOCLDEVFILEOPEN:
	case TLTARGIOCLDEVFILECLOSE:
	case TLTARGIOCLDEVFILETRUNCATE:
	case TLTARGIOCLDEVFILESYNC:
		req.file_spec = __qs_malloc(sizeof(*req.file_spec), M_FILE, M_NOWAIT);
		if (!req.file_spec) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, req.file_spec, sizeof(*req.file_spec))) != 0) {
			free(req.file_spec, M_FILE);
			break;
		}
		if (cmd == TLTARGIOCLDEVFILEOPEN)
			retval = core_itf_file_open(req.file_spec);
		else if (cmd == TLTARGIOCLDEVFILECLOSE)
			retval = core_itf_file_close(req.file_spec);
		else if (cmd == TLTARGIOCLDEVFILETRUNCATE)
			retval = core_itf_file_truncate(req.file_spec);
		else if (cmd == TLTARGIOCLDEVFILESYNC)
			retval = core_itf_file_sync(req.file_spec);
		if (cmd == TLTARGIOCLDEVFILEOPEN) {
			if (retval == 0)
				retval = copyout(req.file_spec, userp, sizeof(*req.file_spec));
			else
				err = copyout(req.file_spec, userp, sizeof(*req.file_spec));
		}
		free(req.file_spec, M_FILE);
		break;
	case TLTARGIOCLDEVBDEVSYNC:
		if ((retval = copyin(userp, &req.bdev_sync, sizeof(req.bdev_sync))) != 0)
			break;
		retval = core_itf_bdev_sync(&req.bdev_sync);
		break;
	case TLTARGIOCLDEVTDEVOPEN:
	case TLTARGIOCLDEVTDEVCLOSE:
		tdev_spec = __qs_malloc(sizeof(*tdev_spec), M_MMAP, M_NOWAIT);
		if (!tdev_spec) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, tdev_spec, sizeof(*tdev_spec))) != 0) {
			free(tdev_spec, M_MMAP);
			break;
		}
		if (cmd == TLTARGIOCLDEVTDEVOPEN)
			retval = core_itf_tdev_open(tdev_spec);
		else if (cmd == TLTARGIOCLDEVTDEVCLOSE)
			retval = core_itf_tdev_close(tdev_spec);
		if (retval == 0)
			retval = copyout(tdev_spec, userp, sizeof(*tdev_spec));
		else
			err = copyout(tdev_spec, userp, sizeof(*tdev_spec));
		free(tdev_spec, M_MMAP);
		break;
	case TLTARGIOCLDEVBDEVOPEN:
	case TLTARGIOCLDEVBDEVCLOSE:
	case TLTARGIOCLDEVBDEVVALIDATE:
		bdev_spec = __qs_malloc(sizeof(*bdev_spec), M_MMAP, M_NOWAIT);
		if (!bdev_spec) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, bdev_spec, sizeof(*bdev_spec))) != 0) {
			free(bdev_spec, M_MMAP);
			break;
		}
		if (cmd == TLTARGIOCLDEVBDEVOPEN)
			retval = core_itf_bdev_open(bdev_spec);
		else if (cmd == TLTARGIOCLDEVBDEVCLOSE)
			retval = core_itf_bdev_close(bdev_spec);
		else if (cmd == TLTARGIOCLDEVBDEVVALIDATE)
			retval = core_itf_bdev_validate(bdev_spec);
		if (retval == 0)
			retval = copyout(bdev_spec, userp, sizeof(*bdev_spec));
		else
			err = copyout(bdev_spec, userp, sizeof(*bdev_spec));
		free(bdev_spec, M_MMAP);
		break;
	case TLTARGIOCLDEVRELAYCMD:
		tape_req = &req.tape_req;
		if ((retval = copyin(userp, tape_req, sizeof(*tape_req))) != 0)
			break;
		pglist = NULL;
		data = NULL;
		if (tape_req->pglist_cnt) {
			pglist = __qs_malloc(sizeof(struct pgdata_spec) * tape_req->pglist_cnt, M_MMAP, M_NOWAIT);
			if (!pglist) {
				retval = -ENOMEM;
				break;
			}

			if ((retval = copyin(tape_req->data, pglist, (sizeof(struct pgdata_spec) * tape_req->pglist_cnt))) != 0) {
				DEBUG_WARN_NEW("Cannot copyin pglist for count %d\n", tape_req->pglist_cnt);
				free(pglist, M_MMAP);
				retval = -EIO;
				break;
			}
		}
		else if (tape_req->dxfer_len) {
			data = __qs_malloc(tape_req->dxfer_len, M_MMAP, M_NOWAIT);
			if (!data) {
				retval = -ENOMEM;
				break;
			}

			if (tape_req->dir == DATA_DIR_TO_TAPE) {
				if ((retval = copyin(tape_req->data, data, tape_req->dxfer_len)) != 0) {
					free(data, M_MMAP);
					retval = -EIO;
					break;
				}
			}
		}
		retval = core_itf_relay_cmnd(tape_req, pglist, data);
		if (retval == 0)
			retval = copyout(tape_req, userp, sizeof(*tape_req));
		else
			err = copyout(tape_req, userp, sizeof(*tape_req));
		if (data)
			free(data, M_MMAP);
		if (pglist)
			free(pglist, M_MMAP);
		break;
	case TLTARGIOCCOMPUNCOMP:
		comp_spec = &req.comp_spec;
		if ((retval = copyin(userp, comp_spec, sizeof(*comp_spec))) != 0)
			break;

		wrkmem = get_wrkmem();
		if (!wrkmem) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(comp_spec->cpages, wrkmem->pages, (sizeof(struct page_spec) * (comp_spec->cpage_count + comp_spec->upage_count)))) != 0) {
			DEBUG_WARN_NEW("Cannot copyin cpages for count %d\n", comp_spec->cpage_count);
			put_wrkmem(wrkmem);
			retval = -EIO;
			break;
		}

		cpages = (struct page_spec *)(wrkmem->pages);
		upages = cpages + comp_spec->cpage_count;
		retval = core_itf_comp_uncomp(comp_spec, cpages, upages, wrkmem);
		if (comp_spec->compress) {
			if (retval == 0)
				retval = copyout(comp_spec, userp, sizeof(*comp_spec));
			else
				err = copyout(comp_spec, userp, sizeof(*comp_spec));
		}
		put_wrkmem(wrkmem);
		break;
	case TLTARGIOCLDEVCTIOSETUP:
		ctio_spec = &req.ctio_spec;
		if ((retval = copyin(userp, ctio_spec, sizeof(*ctio_spec))) != 0)
			break;
		pglist = __qs_malloc(sizeof(struct pgdata_spec) * ctio_spec->pglist_cnt, M_MMAP, M_NOWAIT);
		if (!pglist) {
			retval = -ENOMEM;
			break;
		}
		if ((retval = copyin(ctio_spec->pglist, pglist, (sizeof(struct pgdata_spec) * ctio_spec->pglist_cnt))) != 0) {
			DEBUG_WARN_NEW("Cannot copyin pglist for count %d\n", ctio_spec->pglist_cnt);
			free(pglist, M_COMP);
			retval = -EIO;
			break;
		}
		retval = core_itf_ctio_setup(ctio_spec, pglist);
		free(pglist, M_COMP);
		break;
	case TLTARGIOCUMAINFO:
		uma_spec = &req.uma_spec;
		if ((retval = copyin(userp, uma_spec, sizeof(*uma_spec))) != 0)
			break;
		buffer = __qs_malloc(UMA_INFO_BUFFER_SIZE, M_QUADSTOR, M_NOWAIT);
		if (!buffer) {
			retval = -ENOMEM;
			break;
		}
		retval = ioctl_request(cmd, buffer, UMA_INFO_BUFFER_SIZE);
		if (retval == 0)
			retval = copyout(buffer, uma_spec->buffer, UMA_INFO_BUFFER_SIZE);
		free(buffer, M_QUADSTOR);
		break;
	case TLTARGIOCLDEVBDEVUNMAPSUPPORT:
		bio_spec = &req.bio_spec;
		if ((retval = copyin(userp, bio_spec, sizeof(*bio_spec))) != 0)
			break;
		retval = core_itf_bdev_unmap_support(bio_spec);
		break;
	case TLTARGIOCLDEVBIOUNMAP:
		bio_spec = &req.bio_spec;
		if ((retval = copyin(userp, bio_spec, sizeof(*bio_spec))) != 0)
			break;
		retval = core_itf_bio_unmap(bio_spec);
		break;
	case TLTARGIOCLDEVBIOREQ:
		bio_spec = &req.bio_spec;
		if ((retval = copyin(userp, bio_spec, sizeof(*bio_spec))) != 0)
			break;

		pglist = __qs_malloc(sizeof(struct pgdata_spec) * bio_spec->page_count, M_MMAP, M_NOWAIT);
		if (!pglist) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(bio_spec->pglist, pglist, (sizeof(struct pgdata_spec) * bio_spec->page_count))) != 0) {
			DEBUG_WARN_NEW("Cannot copyin pglist for count %d\n", bio_spec->page_count);
			free(pglist, M_COMP);
			retval = -EIO;
			break;
		}

		retval = core_itf_bio_req(bio_spec, pglist);
		free(pglist, M_COMP);
		break;
	case TLTARGIOCLDEVINIT:
		retval = core_send_init();
		break;
	case TLTARGIOCLDEVDETACH:
		core_itf_detach_interfaces();
		retval = 0;
		break;
	case TLTARGIOCLDEVSOCKREAD:
	case TLTARGIOCLDEVSOCKWRITE:
		sock_data_spec = &req.sock_data_spec;
		if ((retval = copyin(userp, sock_data_spec, sizeof(*sock_data_spec))) != 0)
			break;
		retval = core_itf_sock_rw(sock_data_spec, cmd == TLTARGIOCLDEVSOCKWRITE ? 1 : 0);
		if (retval == 0)
			retval = copyout(sock_data_spec, userp, sizeof(*sock_data_spec));
		else
			err = copyout(sock_data_spec, userp, sizeof(*sock_data_spec));
		break;
	case TLTARGIOCLDEVSOCKCONNECTLOCAL:
		sock_local_spec = &req.sock_local_spec;
		if ((retval = copyin(userp, sock_local_spec, sizeof(*sock_local_spec))) != 0)
			break;
		retval = core_itf_sock_connect_local(sock_local_spec);
		break;
	case TLTARGIOCLDEVSOCKCREATE:
	case TLTARGIOCLDEVSOCKCREATELOCAL:
	case TLTARGIOCLDEVSOCKCONNECT:
	case TLTARGIOCLDEVSOCKSHUTDOWN:
	case TLTARGIOCLDEVSOCKCLOSE:
	case TLTARGIOCLDEVSOCKACCEPT:
	case TLTARGIOCLDEVSOCKBIND:
	case TLTARGIOCLDEVSOCKPOLL:
	case TLTARGIOCLDEVSOCKHASWRITESPACE:
	case TLTARGIOCLDEVSOCKISCONNECTED:
	case TLTARGIOCLDEVSOCKHASREADDATA:
	case TLTARGIOCLDEVSOCKNOPUSH:
		sock_spec = &req.sock_spec;
		if ((retval = copyin(userp, sock_spec, sizeof(*sock_spec))) != 0)
			break;
		if (cmd == TLTARGIOCLDEVSOCKCREATE)
			retval = core_itf_sock_create(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKCREATELOCAL)
			retval = core_itf_sock_create_local(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKCONNECT)
			retval = core_itf_sock_connect(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKSHUTDOWN)
			retval = core_itf_sock_shutdown(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKCLOSE)
			retval = core_itf_sock_close(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKACCEPT)
			retval = core_itf_sock_accept(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKBIND)
			retval = core_itf_sock_bind(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKPOLL)
			retval = core_itf_sock_poll(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKHASWRITESPACE)
			retval = core_itf_sock_has_write_space(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKISCONNECTED)
			retval = core_itf_sock_is_connected(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKHASREADDATA)
			retval = core_itf_sock_has_read_data(sock_spec);
		else if (cmd == TLTARGIOCLDEVSOCKNOPUSH)
			retval = core_itf_sock_nopush(sock_spec);

		if (retval == 0)
			retval = copyout(&req.sock_spec, userp, sizeof(req.sock_spec));
		else
			err = copyout(&req.sock_spec, userp, sizeof(req.sock_spec));
		break;
	case TLTARGIOCLDEVNEW:
	case TLTARGIOCLDEVDELETE:
	case TLTARGIOCLDEVDISABLE:
	case TLTARGIOCLDEVUPDATE:
		if ((retval = copyin(userp, &req.ldev_info, sizeof(req.ldev_info))) != 0)
			break;

		if (cmd == TLTARGIOCLDEVNEW)
			retval = core_itf_device_new(&req.ldev_info);
		else if (cmd == TLTARGIOCLDEVDELETE)
			retval = core_itf_device_delete(&req.ldev_info);
		else if (cmd == TLTARGIOCLDEVDISABLE)
			retval = core_itf_device_disable(&req.ldev_info);
		else if (cmd == TLTARGIOCLDEVUPDATE)
			retval = core_itf_device_update(&req.ldev_info);
		if (retval == 0)
			retval = copyout(&req.ldev_info, userp, sizeof(req.ldev_info));
		else
			err = copyout(&req.ldev_info, userp, sizeof(req.ldev_info));
		break;
	case TLTARGIOCTAPECMD:
		tapecmd = malloc(sizeof(*tapecmd), M_QUADSTOR, M_NOWAIT);
		if (!tapecmd) {
			DEBUG_WARN_NEW("Cannot allocate a new tapecmd\n");
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, tapecmd, sizeof(*tapecmd))) != 0) {
			DEBUG_WARN_NEW("copy in tapecmd failed with retval %d\n", retval);
			free(tapecmd, M_QUADSTOR);
			break;
		}

		ctio = ctio_alloc(tapecmd->device_lunid);
		if (!ctio) {
			free(tapecmd, M_QUADSTOR);
			retval = -ENOMEM;
			break;
		}
		memcpy(ctio->cdb, tapecmd->cdb, 16);
		if (tapecmd->dxfer_out) {
			retval = device_allocate_cmd_buffers(ctio, Q_NOWAIT);
			if (unlikely(retval != 0)) {
				ctio_free_all(ctio);
				free(tapecmd, M_QUADSTOR);
				retval = -ENOMEM;
				break;
			}
			if (tapecmd->dxfer_out != ctio->dxfer_len) {
				DEBUG_WARN_NEW("Mismatch in dxfer out len %d ctio %d\n", tapecmd->dxfer_out, ctio->dxfer_len);
				ctio_free_all(ctio);
				free(tapecmd, M_QUADSTOR);
				retval = -EINVAL;
				break;
			}
			retval = ctio_copy_tape_data(ctio, tapecmd->data, 1, NULL);
			if (unlikely(retval != 0)) {
				DEBUG_WARN_NEW("Ctio copy tape data failed\n");
				ctio_free_all(ctio);
				free(tapecmd, M_QUADSTOR);
				break;
			}
		}
		ctio_queue(ctio);

		if (ctio->dxfer_len && tapecmd->dxfer_in < ctio->dxfer_len) {
			DEBUG_WARN_NEW("Mismatch in dxfer in len %d ctio %d\n", tapecmd->dxfer_in, ctio->dxfer_len);
			retval = -EINVAL;
		}
		else if (ctio->dxfer_len) {
			retval = ctio_copy_tape_data(ctio, tapecmd->data, 0, NULL);
			if (retval != 0)
				DEBUG_WARN_NEW("Ctio copy tape data back failed\n");
		}

		if (ctio->sense_len)
			memcpy(tapecmd->sense, ctio->sense_data, ctio->sense_len);
		tapecmd->dxfer_in = ctio->dxfer_len;
		tapecmd->sense_len = ctio->sense_len;
		tapecmd->scsi_status = ctio->scsi_status;

		if (retval == 0)
			retval = copyout(tapecmd, userp, sizeof(*tapecmd));
		else
			err = copyout(tapecmd, userp, sizeof(*tapecmd));

		ctio_free_all(ctio);
		free(tapecmd, M_QUADSTOR);
		break;
	case TLTARGIOCTAPEWRITE:
	case TLTARGIOCTAPEREAD:
		tapebuf = malloc(sizeof(*tapebuf), M_QUADSTOR, M_NOWAIT);
		if (!tapebuf) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, tapebuf, sizeof(*tapebuf))) != 0) {
			free(tapebuf, M_QUADSTOR);
			break;
		}

		ctio = ctio_alloc(tapebuf->device_lunid);
		if (!ctio) {
			free(tapebuf, M_QUADSTOR);
			retval = -ENOMEM;
			break;
		}

		ctio->cdb[0] = (cmd ==  TLTARGIOCTAPEWRITE) ? WRITE_6 : READ_6;
		ctio->cdb[2] = (tapebuf->dxfer_len >> 16) & 0xFF;
		ctio->cdb[3] = (tapebuf->dxfer_len >> 8) & 0xFF;
		ctio->cdb[4] = (tapebuf->dxfer_len) & 0xFF;

		if (cmd == TLTARGIOCTAPEWRITE) {

			retval = device_allocate_buffers(ctio, Q_NOWAIT);
			if (unlikely(retval != 0)) {
				DEBUG_WARN_NEW("Cannot allocate buffers for %u dxfer_len\n", tapebuf->dxfer_len);
				retval = -ENOMEM;
				free(tapebuf, M_QUADSTOR);
				ctio_free_all(ctio);
				break;
			}

			if (ctio->dxfer_len != tapebuf->dxfer_len) {
				DEBUG_WARN_NEW("Invalid request dxfer len %d cmd dxfer len %d\n", ctio->dxfer_len, tapebuf->dxfer_len);
				retval = -EINVAL;
				free(tapebuf, M_QUADSTOR);
				ctio_free_all(ctio);
				break;
			}

			retval = ctio_copy_tape_data(ctio, tapebuf->data, 1, NULL);
			if (unlikely(retval != 0)) {
				DEBUG_WARN_NEW("Cannot allocate buffers for %u dxfer_len\n", ctio->dxfer_len);
				free(tapebuf, M_QUADSTOR);
				ctio_free_all(ctio);
				break;
			}
		}

		ctio_queue(ctio);

		if (cmd == TLTARGIOCTAPEREAD) {
			retval = ctio_copy_tape_data(ctio, tapebuf->data, 0, &tapebuf->csum);
			if (retval == 0 && ctio->scsi_status == SCSI_STATUS_CHECK_COND) {
				uint8_t sense_key = ctio->sense_data[2] & 0xF;
				uint8_t ili = (ctio->sense_data[2] >> 5) & 0x1;
				uint8_t asc = ctio->sense_data[12];
				uint8_t ascq = ctio->sense_data[13];

				if (sense_key == SSD_KEY_NO_SENSE && !asc && !ascq && ili && ctio->dxfer_len == tapebuf->dxfer_len)
					retval = -ENOMEM;
			}
		}
		else
			retval = 0;
		if (ctio->sense_len)
			memcpy(tapebuf->sense, ctio->sense_data, ctio->sense_len);
		tapebuf->dxfer_len = ctio->dxfer_len;
		tapebuf->sense_len = ctio->sense_len;
		tapebuf->scsi_status = ctio->scsi_status;

		ctio_free_all(ctio);
		if (retval == 0)
			retval = copyout(tapebuf, userp, sizeof(*tapebuf));
		else
			err = copyout(tapebuf, userp, sizeof(*tapebuf));
		free(tapebuf, M_QUADSTOR);
		break;
	case TLTARGIOCMAPVTL:
		ieinfo = __qs_malloc(sizeof(*ieinfo), M_MMAP, M_NOWAIT);
		if (!ieinfo) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, ieinfo, sizeof(*ieinfo))) != 0) {
			free(ieinfo, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, ieinfo, sizeof(*ieinfo));
		if (retval == 0)
			retval = copyout(ieinfo, userp, sizeof(*ieinfo));
		else
			err = copyout(ieinfo, userp, sizeof(*ieinfo));
		free(ieinfo, M_MMAP);
		break;
	case TLTARGIOCMAPSLOT:
	case TLTARGIOCUNMAPSLOT:
		mslot = __qs_malloc(sizeof(*mslot), M_MMAP, M_NOWAIT);
		if (!mslot) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, mslot, sizeof(*mslot))) != 0) {
			free(mslot, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, mslot, sizeof(*mslot));
		if (retval == 0)
			retval = copyout(mslot, userp, sizeof(*mslot));
		else
			err = copyout(mslot, userp, sizeof(*mslot));
		free(mslot, M_MMAP);
		break;
	case TLTARGIOCCOREDEVLOADDONE:
		atomic_set(&core_inited, 1);
		retval = 0;
		break;
	case TLTARGIOCDAEMONSETINFO:
		if (!atomic_read(&core_inited)) {
			retval = -EBUSY;
			break;
		}
		mdaemon_info = &req.mdaemon_info;
		if ((retval = copyin(userp, mdaemon_info, sizeof(*mdaemon_info))) != 0)
			break;
		retval = ioctl_request(cmd, mdaemon_info, sizeof(*mdaemon_info));
		break;
	case TLTARGIOCNODECONFIG:
		node_config = __qs_malloc(sizeof(*node_config), M_MMAP, M_NOWAIT);
		if (!node_config) {
			retval = -ENOMEM;
			break;
		}

		if (cmd == TLTARGIOCNODECONFIG) {
			if ((retval = copyin(userp, node_config, sizeof(*node_config))) != 0) {
				free(node_config, M_MMAP);
				break;
			}
			retval = ioctl_request(cmd, node_config, sizeof(*node_config));
		}
		free(node_config, M_MMAP);
		break;
	case TLTARGIOCADDFCRULE:
	case TLTARGIOCREMOVEFCRULE:
	case TLTARGIOCMODIFYFCRULE:
		if ((retval = copyin(userp, &fc_rule_config, sizeof(fc_rule_config))) != 0)
			break;
		retval = ioctl_request(cmd, &fc_rule_config, sizeof(fc_rule_config));
		break;
	case TLTARGIOCMIRRORVCARTRIDGE:
	case TLTARGIOCMIRRORSTATUS:
	case TLTARGIOCMIRRORCANCEL:
		mirror_config = __qs_malloc(sizeof(*mirror_config), M_MMAP, M_NOWAIT);
		if (!mirror_config) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, mirror_config, sizeof(*mirror_config))) != 0) {
			free(mirror_config, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, mirror_config, sizeof(*mirror_config));
		if (retval == 0)
			retval = copyout(mirror_config, userp, sizeof(*mirror_config));
		else
			err = copyout(mirror_config, userp, sizeof(*mirror_config));
		free(mirror_config, M_MMAP);
		break;
	case TLTARGIOCCLONEVCARTRIDGE:
	case TLTARGIOCCLONESTATUS:
	case TLTARGIOCCLONECANCEL:
		clone_config = __qs_malloc(sizeof(*clone_config), M_MMAP, M_NOWAIT);
		if (!clone_config) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, clone_config, sizeof(*clone_config))) != 0) {
			free(clone_config, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, clone_config, sizeof(*clone_config));
		if (retval == 0)
			retval = copyout(clone_config, userp, sizeof(*clone_config));
		else
			err = copyout(clone_config, userp, sizeof(*clone_config));
		free(clone_config, M_MMAP);
		break;
	case TLTARGIOCNEWBLKDEV:
	case TLTARGIOCDELBLKDEV:
	case TLTARGIOCMODBLKDEV:
	case TLTARGIOCGETBLKDEV:
	case TLTARGIOCUNMAPCONFIG:
	case TLTARGIOCSTARTUNMAP:
	case TLTARGIOCSTOPUNMAP:
		bdev_info = __qs_malloc(sizeof(struct bdev_info), M_MMAP, M_NOWAIT);
		if (!bdev_info) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, bdev_info, sizeof(struct bdev_info))) != 0) {
			free(bdev_info, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, bdev_info, sizeof(*bdev_info));
		if (retval == 0)
			retval = copyout(bdev_info, userp, sizeof(*bdev_info));
		else
			err = copyout(bdev_info, userp, sizeof(*bdev_info));
		free(bdev_info, M_MMAP);
		break;
	case TLTARGIOCVDEVICEDUMPBUFFER:
		vdevicedump = __qs_malloc(sizeof(*vdevicedump), M_MMAP, M_NOWAIT);
		if (!vdevicedump) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, vdevicedump, sizeof(*vdevicedump))) != 0) {
			free(vdevicedump, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, vdevicedump, sizeof(*vdevicedump));
		if (retval == 0)
			retval = copyout(vdevicedump, userp, sizeof(*vdevicedump));
		else
			err = copyout(vdevicedump, userp, sizeof(*vdevicedump));
		free(vdevicedump, M_MMAP);
		break;
	case TLTARGIOCLIBCTL:
		libctl = __qs_malloc(sizeof(*libctl), M_MMAP, M_NOWAIT);
		if (!libctl) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, libctl, sizeof(*libctl))) != 0) {
			free(libctl, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, libctl, sizeof(*libctl));
		if (retval == 0)
			retval = copyout(libctl, userp, sizeof(*libctl));
		else
			err = copyout(libctl, userp, sizeof(*libctl));
		free(libctl, M_MMAP);
		break;
	case TLTARGIOCNEWDEVICE:
	case TLTARGIOCDELETEDEVICE:
	case TLTARGIOCMODDEVICE:
	case TLTARGIOCMODIFYELEMENTCOUNT:
	case TLTARGIOCGETDEVICEINFO:
	case TLTARGIOCLOADDRIVE:
	case TLTARGIOCRESETSTATS:
	case TLTARGIOCDEVICECTL:
	case TLTARGIOCMODIFYSERIAL:
	case TLTARGIOCADDINQUIRY:
	case TLTARGIOCDELETEINQUIRY:
		deviceinfo = __qs_malloc(sizeof(*deviceinfo), M_MMAP, M_NOWAIT);
		if (!deviceinfo) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, deviceinfo, sizeof(*deviceinfo))) != 0) {
			free(deviceinfo, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, deviceinfo, sizeof(*deviceinfo));
		if (retval == 0) {
			if (cmd == TLTARGIOCGETDEVICEINFO)
				copy_device_stats(deviceinfo);
			else if (cmd == TLTARGIOCRESETSTATS)
				reset_device_stats(deviceinfo);
			retval = copyout(deviceinfo, userp, sizeof(*deviceinfo));
		}
		else
			err = copyout(deviceinfo, userp, sizeof(*deviceinfo));
		free(deviceinfo, M_MMAP);
		break;
	case TLTARGIOCNEWVCARTRIDGE:
	case TLTARGIOCLOADVCARTRIDGE:
	case TLTARGIOCDELETEVCARTRIDGE:
	case TLTARGIOCMODIFYVCARTRIDGE:
	case TLTARGIOCGETVCARTRIDGEINFO:
	case TLTARGIOCRELOADEXPORT:
	case TLTARGIOCCLEARREPEXP:
	case TLTARGIOCEXPORTTAPE:
	case TLTARGIOCVAULTVCARTRIDGE:
	case TLTARGIOCEXPORTCLEAR:
	case TLTARGIOCREPLICATIONCLEAR:
		vcartridge = __qs_malloc(sizeof(*vcartridge), M_MMAP, M_NOWAIT);
		if (!vcartridge) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, vcartridge, sizeof(*vcartridge))) != 0) {
			free(vcartridge, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, vcartridge, sizeof(*vcartridge));
		if (retval == 0)
			retval = copyout(vcartridge, userp, sizeof(*vcartridge));
		else
			err = copyout(vcartridge, userp, sizeof(*vcartridge));
		free(vcartridge, M_MMAP);
		break;
	case TLTARGIOCCHECKDISKS:
		retval = ioctl_request(cmd, NULL, 0);
		break;
	case TLTARGIOCLOADDONE:
		retval = ioctl_request(cmd, NULL, 0);
		break;
	case TLTARGIOCQLOADDONE:
		retval = ioctl_request(cmd, NULL, 0);
		if (retval == 0)
			retval = ldev_qload_done();
		break;
	case TLTARGIOCUNLOAD:
		atomic_set(&itf_disabled, 1);
		retval = ioctl_request(cmd, NULL, 0);
		break;
	case TLTARGIOCDDTABLELOADSTATUS:
	case TLTARGIOCDDTABLEALIGN:
	case TLTARGIOCDDTABLEFILL:
		retval = ioctl_request(cmd, NULL, 0);
		break;
	case TLTARGIOCADDGROUP:
	case TLTARGIOCDELETEGROUP:
	case TLTARGIOCRENAMEGROUP:
		group_conf = __qs_malloc(sizeof(*group_conf), M_MMAP, M_NOWAIT);
		if (!group_conf) {
			retval = -ENOMEM;
			break;
		}

		if ((retval = copyin(userp, group_conf, sizeof(*group_conf))) != 0) {
			free(group_conf, M_MMAP);
			break;
		}

		retval = ioctl_request(cmd, group_conf, sizeof(*group_conf));
		free(group_conf, M_MMAP);
		break;
	default:
		break;
	}

	return retval;
}

static int
have_bio_msg(void)
{
	struct bio_msg *msg;
	int idx = (msg_idx - 1);

	if (idx < 0)
		idx = BIO_MAX_MSG - 1;

	msg = (struct bio_msg *)(msg_addr + (idx  * sizeof(*msg)));
	return msg->inuse;
}

static unsigned int
coreitf_poll(struct file *filp, poll_table *wait)
{
	struct tdevice *tdevice;
	struct bdevice *bdevice;
	unsigned int ret = 0;
	unsigned long flags;

	tdevice = filp->private_data;
	if (tdevice->type == TYPE_BDEVICE) {
		bdevice = filp->private_data;
		if (atomic_read(&bdevice->disabled)) {
			ret = POLLHUP;
			filp->private_data = NULL;
			return ret;
		}
		spin_lock_irqsave(&bio_msg_lock, flags);
		if (have_bio_msg())
			ret |= POLLIN | POLLRDNORM;
		else
			poll_wait(filp, &bio_msg_wait, wait);
		spin_unlock_irqrestore(&bio_msg_lock, flags);
	}
	else if (tdevice && !atomic_read(&tdevice->disabled)) {
		spin_lock_irqsave(&tdevice->tdevice_lock, flags);
		if (!atomic_read(&tdevice->pending_msg))
			poll_wait(filp, &tdevice->wait, wait);
		else {
			atomic_dec(&tdevice->pending_msg);
			ret |= POLLIN | POLLRDNORM;
		}
		spin_unlock_irqrestore(&tdevice->tdevice_lock, flags);
	}
	else if (tdevice) {
		ret = POLLHUP;
		tdevice_put(tdevice);
		filp->private_data = NULL;
	}
	return ret;
}

static struct file_operations coreitf_fops = {
	.owner = THIS_MODULE,
	.mmap = coreitf_mmap,
	.llseek = noop_llseek,
	.poll = coreitf_poll,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0))
	.compat_ioctl = coreitf_ioctl,
	.unlocked_ioctl = coreitf_ioctl,
#else
	.ioctl = coreitf_ioctl,
#endif
};

static int dev_major;
static kproc_t *bio_tsk;

static struct bio_priv *
get_next_bio_priv(void)
{
	struct bio_priv *priv;
	unsigned long flags;

	chan_lock_intr(&bio_chan, &flags);
	priv = TAILQ_FIRST(&bio_priv_list);
	if (priv)
		TAILQ_REMOVE(&bio_priv_list, priv, b_entry);
	chan_unlock_intr(&bio_chan, &flags);
	return priv;
}

static void
process_bio_priv_list(void)
{
	struct bio_msg *msg;
	struct bio_priv *priv;
	unsigned long flags;

	while ((priv = get_next_bio_priv()) != NULL) {
		msg = (struct bio_msg *)(msg_addr + (msg_idx  * sizeof(*msg)));
		while (msg->inuse)
			msleep(1);
		spin_lock_irqsave(&bio_msg_lock, flags);
		msg->bio_addr = priv->bio_addr;
		msg->error = atomic_read(&priv->error);
		mb();
		msg->inuse = 1;
		msg_idx++;
		if (msg_idx == BIO_MAX_MSG)
			msg_idx = 0;
		wake_up_interruptible(&bio_msg_wait);
		spin_unlock_irqrestore(&bio_msg_lock, flags);
		kfree(priv);
	}
}

static int
bio_complete_thread(void *data)
{
	set_user_nice(current, -20);

	for (;;) {
		wait_on_chan_interruptible(bio_chan, !TAILQ_EMPTY(&bio_priv_list) || kthread_should_stop());
		process_bio_priv_list();
		if (kthread_should_stop())
			break;
	}
	return 0;
}

static int
coreitf_init(void)
{
	struct wrkmem *wrkmem;
	int retval;
	int i;

	sx_init(&cbs_lock, "cbs lck");
	sx_init(&bdevs_lock, "bdevs lck");
	sx_init(&socks_lock, "socks lck");
	sx_init(&comm_socks_lock, "comm socks lck");
	sx_init(&tdevs_lock, "tdevs lck");
	spin_lock_init(&wrkmem_lock);
	TAILQ_INIT(&wrkmem_list);
	LIST_INIT(&cbs_list);
	spin_lock_init(&bio_msg_lock);
	wait_chan_init(&bio_chan, "bio chan");
	init_waitqueue_head(&bio_msg_wait);
	init_waitqueue_head(&comm_sock_wait);
	mmap_table_init();

	retval = init_caches();
	if (unlikely(retval != 0)) {
		return -(EIO);
	}

	dev_major = register_chrdev(0, TL_DEV_NAME, &coreitf_fops);
	if (unlikely(dev_major < 0)) {
		exit_caches();
		return -(EIO);
	}

	bio_tsk = kthread_run(bio_complete_thread, NULL, "biocomp");
	if (IS_ERR(bio_tsk)) {
		unregister_chrdev(dev_major, TL_DEV_NAME);
		core_send_exit();
		mmap_table_exit();
		exit_caches();
		return -(EIO);
	}

	/* Init a few compression buffers */
	for (i = 0; i < 64; i++) {
		wrkmem = alloc_wrkmem();
		if (!wrkmem)
			break;
		TAILQ_INSERT_TAIL(&wrkmem_list, wrkmem, w_list);
		atomic_inc(&wrkmem_count);
	}

	return 0;
}

static void
coreitf_exit(void)
{
	struct wrkmem *wrkmem;
	int i;

	if (bio_tsk)
		kthread_stop(bio_tsk);

	while ((wrkmem = TAILQ_FIRST(&wrkmem_list)) != NULL) {
		TAILQ_REMOVE(&wrkmem_list, wrkmem, w_list);
		free_wrkmem(wrkmem);
	}
	for (i = 0; i < TL_MAX_KFDS; i++) {
		if (!bdevs[i])
			continue;
		DEBUG_INFO_NEW("close bdev at %d\n", i);
		core_itf_iodev_close(bdevs[i]);
	}
	for (i = 0; i < TL_MAX_KFDS; i++) {
		if (!tdevs[i])
			continue;
		DEBUG_INFO_NEW("close tdev at %d\n", i);
		scsi_device_put(tdevs[i]);
	}
	for (i = 0; i < TL_MAX_SOCKS; i++) {
		if (!socks[i])
			continue;
		DEBUG_INFO_NEW("close sock at %d\n", i);
		sys_sock_close(socks[i], 0);
		sys_sock_free(socks[i]);
	}
	unregister_chrdev(dev_major, TL_DEV_NAME);
	core_send_exit();
	mmap_table_exit();
	if (bdevice)
		kfree(bdevice);
	exit_caches();
}

MODULE_AUTHOR("Shivaram Upadhyayula, QUADStor Systems");
MODULE_LICENSE("GPL");
module_init (coreitf_init);
module_exit (coreitf_exit);
EXPORT_SYMBOL_GPL(vtdevice_register_interface);
EXPORT_SYMBOL_GPL(vtdevice_unregister_interface);
