/* Socket related stuff */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,15,0))
static void sys_sock_data_ready(struct sock *sk);
#else
static void sys_sock_data_ready(struct sock *sk, int count);
#endif
static void sys_sock_write_space(struct sock *sk);
static void sys_sock_state_change(struct sock *sk);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
int
sys_sock_has_read_data(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;
	int retval;

	if (sys_sock->listen)
		return inet_csk_listen_poll(sock->sk);
	if (sys_sock->usock)
		retval = unix_inq_len(sock->sk);
	else
		retval = tcp_peek_len(sock);
	return retval;
}
#else
int
sys_sock_has_read_data(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;
	mm_segment_t oldfs;
	int avail = 0;
	int retval;

	if (sys_sock->listen)
		return inet_csk_listen_poll(sock->sk);
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	retval = sock->ops->ioctl(sock, SIOCINQ, (unsigned long)&avail);
	set_fs(oldfs);
	return (retval >= 0) ? avail : 0;
}
#endif

int
sys_sock_has_write_space(sock_t *sys_sock)
{
	return (sk_stream_wspace(sys_sock->sock->sk) >= sk_stream_min_wspace(sys_sock->sock->sk));
}

int
sys_sock_is_connected(sock_t *sys_sock)
{
	return (sys_sock->sock->sk->sk_state == TCP_ESTABLISHED);
}

int
sys_sock_read(sock_t *sys_sock, void *buf, int len)
{
	struct msghdr msg;
	struct kvec iov;
	int retval;

	iov.iov_base = buf;
	iov.iov_len = len;

	memset(&msg, 0, sizeof(msg));
	msg.msg_flags = MSG_DONTWAIT | MSG_NOSIGNAL;
	retval = kernel_recvmsg(sys_sock->sock, &msg, &iov, 1, len, MSG_DONTWAIT | MSG_NOSIGNAL);
	if (retval > 0)
		return retval;
	else if (retval == -EAGAIN || retval == -EINTR)
		return 0;
	else
		return retval;
}

void
sys_sock_free(sock_t *sys_sock)
{
	sock_release(sys_sock->sock);
	__node_sock_free(sys_sock);
}

static void
sock_deactivate(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;

	write_lock_bh(&sock->sk->sk_callback_lock);

	/* Restore callbacks */
	if (sys_sock->state_change) {
		sock->sk->sk_state_change = sys_sock->state_change;
		sys_sock->state_change = NULL;
	}

	if (sys_sock->data_ready) {
		sock->sk->sk_data_ready = sys_sock->data_ready;
		sys_sock->data_ready = NULL;
	}

	if (sys_sock->write_space) {
		sock->sk->sk_write_space = sys_sock->write_space;
		sys_sock->write_space = NULL;
	}

	sock->sk->sk_user_data = NULL;
	write_unlock_bh(&sock->sk->sk_callback_lock);
}

int
sys_sock_shutdown(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;
	return kernel_sock_shutdown(sock, SHUT_RDWR);
}

void
sys_sock_close(sock_t *sys_sock, int linger)
{
	struct socket *sock = sys_sock->sock;

	if (!linger)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
		sock_no_linger(sock->sk);
#else
		kernel_setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *)&linger, sizeof(linger));
#endif

	kernel_sock_shutdown(sock, SHUT_RDWR);
	sock_deactivate(sys_sock);
}

int
sys_sock_write_page(sock_t *sys_sock, pagestruct_t *page, int offset, int len)
{
	int retval;

	retval =  qs_sock_write_page(sys_sock->sock, page, offset, len, 0);
	if (retval > 0)
		return retval;
	else if (retval == -EAGAIN || retval == -EINTR)
		return 0;
	else
		return retval;
}

int
sys_sock_write(sock_t *sys_sock, void *buf, int len)
{
	int retval;

	retval = qs_sock_write(sys_sock->sock, buf, len, MSG_DONTWAIT | MSG_NOSIGNAL);
	if (retval > 0)
		return retval;
	else if (retval == -EAGAIN || retval == -EINTR)
		return 0;
	else
		return retval;
}

static sock_t *
sys_sock_create_lite(void *priv)
{
	sock_t *sys_sock;
	int retval;

	sys_sock = __node_sock_alloc(priv);

	retval = sock_create_lite(AF_INET, SOCK_STREAM, IPPROTO_TCP, &sys_sock->sock);
	if (unlikely(retval < 0)) {
		__node_sock_free(sys_sock);
		return NULL;
	}
	return sys_sock;
}

sock_t *
__sys_sock_create(void *priv, int domain)
{
	sock_t *sys_sock;
	int retval;

	sys_sock = __node_sock_alloc(priv);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,2,0))
	retval = sock_create_kern(&init_net, domain, SOCK_STREAM, domain == AF_LOCAL ? 0: IPPROTO_TCP, &sys_sock->sock);
#else
	retval = sock_create_kern(domain, SOCK_STREAM, domain == AF_LOCAL ? 0: IPPROTO_TCP, &sys_sock->sock);
#endif
	if (unlikely(retval < 0)) {
		DEBUG_WARN_NEW("Cannot create sock retval %d\n", retval);
		__node_sock_free(sys_sock);
		return NULL;
	}
	sys_sock->priv = priv;
	return sys_sock;
}

sock_t *
sys_sock_create(void *priv)
{
	return __sys_sock_create(priv, AF_INET);
}

sock_t *
sys_sock_create_local(void *priv)
{
	return __sys_sock_create(priv, AF_LOCAL);
}

void
sock_activate(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;

	write_lock_bh(&sock->sk->sk_callback_lock);
	sock->sk->sk_user_data = sys_sock;
	sock->sk->sk_allocation = GFP_NOFS;

	/* Save callbacks */
	sys_sock->state_change = sock->sk->sk_state_change;
	sys_sock->data_ready = sock->sk->sk_data_ready;
	sys_sock->write_space = sock->sk->sk_write_space;

	/* Set new callbacks */
	sock->sk->sk_state_change = sys_sock_state_change;
	sock->sk->sk_write_space = sys_sock_write_space;
	sock->sk->sk_data_ready = sys_sock_data_ready;
	write_unlock_bh(&sock->sk->sk_callback_lock);
}

void
sys_sock_nopush(sock_t *sys_sock, int set)
{
	if (sys_sock->usock)
		return;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
	tcp_sock_set_cork(sys_sock->sock->sk, set);
#else
	kernel_setsockopt(sys_sock->sock, SOL_TCP, TCP_CORK, (void *)&set, sizeof(set));
#endif
}

sock_t *
sys_sock_accept(sock_t *sys_sock, void *priv, int *error, uint32_t *ipaddr)
{
	sock_t *new_syssock;
	struct socket *sock = sys_sock->sock;
	struct socket *newsock;
	struct sockaddr_in *sinaddr;
	int retval;

	new_syssock = sys_sock_create_lite(priv);
	new_syssock->usock = sys_sock->usock;
	newsock = new_syssock->sock;
	newsock->type = sock->type;
	newsock->ops = sock->ops;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0))
	retval = sock->ops->accept(sock, newsock, O_NONBLOCK, false);
#else
	retval = sock->ops->accept(sock, newsock, O_NONBLOCK);
#endif
	if (retval < 0) {
		*error = retval;
		sock_release(newsock);
		__node_sock_free(new_syssock);
		return NULL;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,18,0))
	retval = newsock->ops->getname(newsock, (struct sockaddr *)&new_syssock->saddr, 2);
#else
	retval = newsock->ops->getname(newsock, (struct sockaddr *)&new_syssock->saddr, &new_syssock->saddr_len, 2);
#endif
	if (retval < 0) {
		*error = retval;
		sock_release(newsock);
		__node_sock_free(new_syssock);
		return NULL;
        }

	sock_activate(new_syssock);
	if (newsock->sk->sk_state == TCP_ESTABLISHED) {
		sys_sock_state_change(newsock->sk);
	}

	if (!new_syssock->usock) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
		tcp_sock_set_nodelay(newsock->sk);
#else
		int disabled = 1;
		kernel_setsockopt(newsock, SOL_TCP, TCP_NODELAY, (void *)&disabled, sizeof(disabled));
#endif
	}
	sinaddr = (struct sockaddr_in *)(&new_syssock->saddr);
	*ipaddr = sinaddr->sin_addr.s_addr;
	*error = 0;
	return new_syssock;
}

int
sys_sock_bind(sock_t *sys_sock, uint32_t addr, uint16_t port)
{
	struct socket *sock = sys_sock->sock;
	struct sockaddr_in saddr_in;
	int retval;

	memset(&saddr_in, 0, sizeof(saddr_in));
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = htons(port);
	saddr_in.sin_addr.s_addr = addr; 

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
	sock_set_reuseaddr(sock->sk);
#else
	{
		int reuse = 1;
		kernel_setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&reuse, sizeof(reuse));
	}
#endif
	retval = sock->ops->bind(sock, (struct sockaddr *)&saddr_in, sizeof(saddr_in));
	if (retval < 0)
		return retval;

	retval = sock->ops->listen(sock, 1024);
	if (retval < 0)
		return retval;

	sys_sock->listen = 1;
	sock_activate(sys_sock);
	return 0;
}

int
sys_sock_connect_local(sock_t *sys_sock, char *path)
{
	struct socket *sock = sys_sock->sock;
	struct sockaddr_un un_addr;
	size_t size;
	int retval;

	memset(&un_addr, 0, sizeof(un_addr));
	un_addr.sun_family = AF_LOCAL;
	strcpy((char *)un_addr.sun_path, path);
	sys_sock->usock = 1;

	sock_activate(sys_sock);
	size = (offsetof (struct sockaddr_un, sun_path) + strlen (un_addr.sun_path));
	retval = sock->ops->connect(sock, (struct sockaddr *)&un_addr, size, O_NONBLOCK);
	if (retval != 0 && retval != -EINPROGRESS)
		return retval;
	return 0;
}

int
sys_sock_connect(sock_t *sys_sock, uint32_t addr, uint32_t local_addr, uint16_t port)
{
	struct socket *sock = sys_sock->sock;
	struct sockaddr_in saddr_in;
	int retval;

	if (!local_addr || (local_addr == addr))
		goto skip_bind;

	memset(&saddr_in, 0, sizeof(saddr_in));
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_addr.s_addr = local_addr; 

	retval = sock->ops->bind(sock, (struct sockaddr *)&saddr_in, sizeof(saddr_in));
	if (retval < 0)
		return retval;

skip_bind:
	memset(&saddr_in, 0, sizeof(saddr_in));
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = htons(port);
	saddr_in.sin_addr.s_addr = addr; 
	sock_activate(sys_sock);
	retval = sock->ops->connect(sock, (struct sockaddr *)&saddr_in, sizeof(saddr_in), O_NONBLOCK);
	if (retval != 0 && retval != -EINPROGRESS)
		return retval;

	if (!sys_sock->usock) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
		tcp_sock_set_nodelay(sock->sk);
#else
		int disabled = 1;
		kernel_setsockopt(sock, SOL_TCP, TCP_NODELAY, (void *)&disabled, sizeof(disabled));
#endif
	}
	return 0;
}

static atomic_t vmap_count;

void *
vm_pg_map(pagestruct_t **pp, int pg_count)
{
	void *addr;
	int tries = 0;

	atomic_inc(&vmap_count);
	if (atomic_read(&vmap_count) > 2048) {
		atomic_set(&vmap_count, 0);
		vm_unmap_aliases(); /* Unmap aliases periodically */
	}

retry:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)) ||  (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8 && RHEL_MINOR >= 5)
	addr = vm_map_ram(pp, pg_count, -1);
#else
	addr = vm_map_ram(pp, pg_count, -1, PAGE_KERNEL);
#endif
	if (addr || ++tries > 1)
		return addr;
	vm_unmap_aliases();
	goto retry;
}

void
vm_pg_unmap(void *maddr, int pg_count)
{
	vm_unmap_ram(maddr, pg_count);
}

static int
ctio_timeout_msecs(struct qsio_scsiio *ctio)
{
	switch (ctio->cdb[0]) {
	case SPACE:
	case SPACE_16:
	case LOCATE:
	case LOCATE_16:
	case WRITE_FILEMARKS:
		return (600 * 1000);
	default:
		return (300 * 1000);
	}
}

static int
ctio_direction(struct qsio_scsiio *ctio)
{
	switch (ctio->cdb[0]) {
	case READ_6:
	case READ_16:
	case MODE_SENSE_6:
	case MODE_SENSE_10:
	case READ_ATTRIBUTE:
	case READ_BUFFER:
	case RECEIVE_DIAGNOSTIC_RESULTS:
	case SECURITY_PROTOCOL_IN:
	case PERSISTENT_RESERVE_IN:
	case FORMAT_MEDIUM:
	case READ_BLOCK_LIMITS:
	case REQUEST_SENSE:
	case LOG_SENSE:
	case REPORT_DENSITY_SUPPORT:
	case READ_POSITION:
		return DMA_FROM_DEVICE;
	case LOG_SELECT: 
	case MODE_SELECT_6:
	case MODE_SELECT_10:
	case WRITE_ATTRIBUTE:
	case WRITE_DYNAMIC_RUNTIME_ATTRIBUTE:
	case PERSISTENT_RESERVE_OUT:
	case WRITE_6:
	case WRITE_16:
	case SEND_DIAGNOSTIC:
		return DMA_TO_DEVICE;
	default:
		return DMA_NONE;
	}
}

static void
copy_ctio_buffer(struct qsio_scsiio *ctio, uint8_t *buffer, int read, int usr)
{
	struct pgdata *pglist, *pgdata;
	int i, offset, min_len, orig_pglist_cnt;

	DEBUG_INFO("ctio pglist cnt %d buffer %p data_ptr %p dxfer_len %d read %d\n", ctio->pglist_cnt, buffer, ctio->data_ptr, ctio->dxfer_len, read);

	if (!ctio->pglist_cnt) {
		if (read)
			memcpy(ctio->data_ptr, buffer, ctio->dxfer_len);
		else
			memcpy(buffer, ctio->data_ptr, ctio->dxfer_len);
		return;
	}

	pglist = (struct pgdata *)(ctio->data_ptr);
	offset = 0;
	for (i = 0; i < ctio->pglist_cnt; i++) {
		pgdata = &pglist[i];
		min_len = min_t(int, ctio->dxfer_len - offset, pgdata->pg_len);
		if (read)
			memcpy(page_address(pgdata->page)+pgdata->pg_offset, buffer+offset, min_len);
		else
			memcpy(buffer+offset, page_address(pgdata->page)+pgdata->pg_offset, min_len);
		DEBUG_BUG_ON(pgdata->pg_len != min_len);
		offset += min_len;
		pgdata->pg_len = min_len;
		if (offset == ctio->dxfer_len)
			break;
	}
	orig_pglist_cnt = ctio->pglist_cnt;
	ctio->pglist_cnt = (i + 1);

	if (usr)
		return;

	for (i = ctio->pglist_cnt; i < orig_pglist_cnt; i++) {
		pgdata = &pglist[i];
		__free_page(pgdata->page);
		pgdata->page = NULL;
	}
}

int
__relay_cmnd(void *mdevice, struct qsio_scsiio *ctio, int *ret_resid, int usr)
{
	struct scsi_device *device = mdevice;
	uint8_t *buffer = NULL;
	unsigned char sense[SCSI_SENSE_BUFFERSIZE];
	int retval, resid = 0, free_buffer = 0;
	int sense_len, sense_offset;

	DEBUG_INFO("ctio dxfer len %d data ptr %p pglist cnt %d\n", ctio->dxfer_len, ctio->data_ptr, ctio->pglist_cnt);
	if (!ctio->dxfer_len || (ctio->data_ptr && !ctio->pglist_cnt)) {
		buffer = ctio->data_ptr;
		goto sendreq;
	}

	buffer = kmalloc(ctio->dxfer_len, GFP_KERNEL);
	if (unlikely(!buffer)) {
		DEBUG_WARN_NEW("Cannot allocate for len %d\n", ctio->dxfer_len);
		return -1;
	}
	free_buffer = 1;

	if (ctio_direction(ctio) == DMA_TO_DEVICE)
		copy_ctio_buffer(ctio, buffer, 0, usr);
sendreq:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,5,0)) || (defined(CONFIG_SUSE_KERNEL) && LINUX_VERSION_CODE >= KERNEL_VERSION(6, 4, 0))
	{
		struct scsi_sense_hdr sshdr;
		const struct scsi_exec_args exec_args = {
			.sshdr = &sshdr,
			.sense = sense,
			.sense_len = sizeof(sense),
			.resid = &resid,
		};
		int direction;

		switch(ctio_direction(ctio)) {
		case DMA_FROM_DEVICE:
			direction = REQ_OP_DRV_IN;
			break;
		case DMA_TO_DEVICE:
			direction = REQ_OP_DRV_OUT;
			break;
		default:
			direction = REQ_OP_DRV_IN;
		}
		retval = scsi_execute_cmd(device, ctio->cdb, direction, buffer, ctio->dxfer_len, msecs_to_jiffies(ctio_timeout_msecs(ctio)), 1, &exec_args);
		if (retval < 0) {
			DEBUG_WARN_NEW("failed to execute command %x retval %d\n", ctio->cdb[0], retval);
			return -1;
		}

		if (scsi_sense_valid(&sshdr)) {
			sense_len = sense[7] + 8;
			sense_offset = (ctio->init_int != TARGET_INT_ISCSI) ? 0 : 2;
			DEBUG_INFO("fill sense for sense len %d sense offset %x\n", sense_len, sense_offset);
			ctio_allocate_sense(ctio, sense_len);
			memcpy(ctio->sense_data+sense_offset, sense, sense_len);
		}
		if (ctio->dxfer_len && ctio_direction(ctio) == DMA_FROM_DEVICE && ctio->pglist_cnt)
			copy_ctio_buffer(ctio, buffer, 1, usr);
		if (ret_resid)
			*ret_resid = resid;
		if (free_buffer)
			kfree(buffer);
	}
#else
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0))
	retval = scsi_execute(device, ctio->cdb, ctio_direction(ctio), buffer, ctio->dxfer_len, sense, NULL, msecs_to_jiffies(ctio_timeout_msecs(ctio)), 1, 0, 0, &resid);
#else
	retval = scsi_execute(device, ctio->cdb, ctio_direction(ctio), buffer, ctio->dxfer_len, sense, msecs_to_jiffies(ctio_timeout_msecs(ctio)), 1, 0, &resid);
#endif
	DEBUG_INFO("ctio cdb %x dxfer_len %d resid %d retval %d\n", ctio->cdb[0], ctio->dxfer_len, resid, retval);
	if (resid <= ctio->dxfer_len && ctio_direction(ctio) == DMA_FROM_DEVICE) {
		int avail = (ctio->dxfer_len - resid);

		if (!avail)
			relay_ctio_free_data(ctio);
		else
			ctio->dxfer_len = avail;
	}

	if (driver_byte(retval) == DRIVER_SENSE) {
		sense_len = sense[7] + 8;
		sense_offset = (ctio->init_int != TARGET_INT_ISCSI) ? 0 : 2;
		DEBUG_INFO("fill sense for sense len %d sense offset %x\n", sense_len, sense_offset);
		ctio_allocate_sense(ctio, sense_len);
		memcpy(ctio->sense_data+sense_offset, sense, sense_len);
	}

	if (ctio->dxfer_len && ctio_direction(ctio) == DMA_FROM_DEVICE && ctio->pglist_cnt)
		copy_ctio_buffer(ctio, buffer, 1, usr);
	if (ret_resid)
		*ret_resid = resid;
	if (free_buffer)
		kfree(buffer);
	if (retval != 0 && driver_byte(retval) != DRIVER_SENSE) {
		DEBUG_WARN_NEW("failed to execute command %x retval %d\n", ctio->cdb[0], retval);
		return -1;
	}
#endif
	return 0;
}
 
int
relay_cmnd(void *mdevice, struct qsio_scsiio *ctio, int *ret_resid)
{
	return __relay_cmnd(mdevice, ctio, ret_resid, 0);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,11,0))
int
validate_block_device(const char *devpath,  iodev_t *b_dev)
{
	dev_t lookup;
	int error;

	error = lookup_bdev(devpath, &lookup);
	if (error) {
		DEBUG_WARN_NEW("Cannot lookup %s\n", devpath);
		return -1;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
	if (file_bdev(b_dev)->bd_dev == lookup)
#else
	if (b_dev->bd_dev == lookup)
#endif
		return 1;
	else {
		DEBUG_INFO_NEW("devpath %s changed\n", devpath);
		return 0;
	}
}
#else
int
validate_block_device(const char *devpath,  iodev_t *b_dev)
{
	iodev_t *lookup;

	lookup = lookup_bdev(devpath);
	if (unlikely(IS_ERR(b_dev))) {
		DEBUG_WARN_NEW("Cannot lookup %s\n", devpath);
		return -1;
	}
	bdput(lookup);
	DEBUG_INFO_NEW("devpath %s lookup %p b_dev %p\n", devpath, lookup, b_dev);
	if (lookup == b_dev)
		return 1;
	else {
		DEBUG_INFO_NEW("devpath %s changed to lookup %p from b_dev %p\n", devpath, lookup, b_dev);
		return 0;
	}
}
#endif

int
ctio_copy_tape_data(struct qsio_scsiio *ctio, uint8_t *buffer, int read_from_user, uint32_t *csum)
{
	struct pgdata *pglist, *pgdata;
	int i, offset, min_len, retval;

	if (!ctio->pglist_cnt) {
		if (read_from_user)
			return copy_from_user(ctio->data_ptr, buffer, ctio->dxfer_len);
		else
			return copy_to_user(buffer, ctio->data_ptr, ctio->dxfer_len);
	}

	if (csum)
		*csum = 0xffffffffU;
	pglist = (struct pgdata *)(ctio->data_ptr);
	offset = 0;
	for (i = 0; i < ctio->pglist_cnt; i++) {
		pgdata = &pglist[i];
		min_len = min_t(int, ctio->dxfer_len - offset, pgdata->pg_len);
		if (read_from_user)
			retval = copy_from_user(page_address(pgdata->page)+pgdata->pg_offset, buffer+offset, min_len);
		else {
			retval = copy_to_user(buffer+offset, page_address(pgdata->page)+pgdata->pg_offset, min_len);
			if (csum)
				*csum = __calc_crc32(*csum, page_address(pgdata->page)+pgdata->pg_offset, min_len);
		}
		if (retval != 0)
			return -EFAULT;
		offset += min_len;
		pgdata->pg_len = min_len;
		if (offset == ctio->dxfer_len)
			break;
	}
	return 0;
}

